/***************************************************************************
 *   Copyright (C) 2006 by Holger Gerth                                      *
 *   holger.gerth@informatik.tu-chemnitz.de                                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "dtgrid.h"
#include "citeratornd.h"
#include "locatornd.h"

CDtGrid::CDtGrid()
{

    this->projND = NULL;
    this->parent = NULL;
    dimension = 1;

    accessTimeRandomAccess = 0;
    accessAverageTimeRandomAccess = 0;
    accessTimeSequentialAccess = 0;

}


CDtGrid::CDtGrid( unsigned int dimension, CDtGrid *parent )
{
    this->dimension = dimension;
    this->parent = parent;

    if ( dimension > 1 ) {

        this->projND = new CDtGrid( dimension-1, this );

    }
    else {

        this->projND = NULL;

    }

    accessTimeRandomAccess = 0;
    accessAverageTimeRandomAccess = 0;
    accessTimeSequentialAccess = 0;

}


CDtGrid::~CDtGrid()
{
    if (projND != NULL) delete projND;
}


/*!
    \fn CDtGrid::convert_to_dtgrid( CVoxelObject3D* vo )
 */
// converts a voxel object into a DT Grid level set
bool CDtGrid::convert_to_dtgrid( CVoxelObject3D* vo ) {

    if (this->dimension != 3) {
        cerr << "DT Grid has false dimension - need to be 3!" << endl;
        return false;
    }
    else {

        CVoxelObject3D* tmp_vo = new CVoxelObject3D();
        bool returnValue;

        tmp_vo->Copy( vo );

        returnValue = this->create_dtgrid_from_voxel_object( tmp_vo );

        delete tmp_vo;

//         this->printDataStructure();

        return returnValue;
    }

}

/*!
    \fn CDtGrid::randomAccess( vector<int> coordinates )
 */
triple CDtGrid::randomAccess( vector<int> coordinates )
{
    triple returnValue;
    int kMin,kMax;
    int coord = coordinates.back();                         // get coordinate
    coordinates.pop_back();                                 // remove this coordinate from the vector

    if ( this->dimension == 1 ) {
        kMin = 0;
        kMax = this->nCoord.size() -1;
    }
    else {
        returnValue = this->projND->randomAccess( coordinates );
        if ( returnValue.inside == false ) {
            returnValue.inside = false;
            returnValue.iValue = 0;
            returnValue.value = this->nCoord.size();
            return returnValue;
         }

        kMin = this->projND->valueIndexPair[returnValue.iValue].iCoord;
        if ( returnValue.iValue == (this->projND->valueIndexPair.size()-1) )
            kMax = this->nCoord.size() -1;
        else
            kMax = this->projND->valueIndexPair[returnValue.iValue +1].iCoord -1;
    }

    if( (coord < this->nCoord.at( kMin )) || (coord > this->nCoord.at( kMax )) ) {
        returnValue.inside = false;
        returnValue.iValue = 0;
        returnValue.value = this->nCoord.size();
        return returnValue;
    }

    // binary search for coord between kMin and kMax
    int k = this->accessBinSearch( coord, kMin, kMax-1 );

    if ( coord <= this->nCoord.at( k+1 ) ) {
        returnValue.inside = true;
        returnValue.iValue = this->acc.at(k>>1) + coord - this->nCoord.at(k);
        if (this->parent == NULL) returnValue.value = this->value.at( returnValue.iValue );
        return returnValue;
    }
    else {
        returnValue.inside = false;
        returnValue.iValue = 0;
//         returnValue.value = -(this->value.at( this->acc.at(k>>1) ) * this->nCoord.size());
        return returnValue;
    }

    return returnValue;

}

triple CDtGrid::randomAccess( int X, int Y, int Z )
{

    vector<int>     coordinate;
    coordinate.push_back(X);
    coordinate.push_back(Y);
    coordinate.push_back(Z);

    return this->randomAccess( coordinate );

}

/*!
    \fn CDtGrid::sequentialAccess( vector<int> coordinates )
 */
// return all defined grid points
vector<coordinate_value> CDtGrid::sequentialAccess()
{
    coordinate_value tmp;
    vector<coordinate_value> returnValue;

    CLocatorND* loc = new CLocatorND( this->dimension, this );

    CIteratorND* it = new CIteratorND( this->dimension, this );
    it->value = this->value.at(0);

    // iterate the locator to the last point
    do {
        tmp.coordinates = loc->getCoordinate();
        tmp.value = it->value;

        returnValue.push_back( tmp );
    } while ( it->increment( loc ) == true );

    delete loc;
    delete it;

    return returnValue;
}

/*!
    \fn CDtGrid::push( vector<int> coordinates, double value )
 */
// the push algorithm
void CDtGrid::push( vector<int> coordinates, double value )
{
    if ( this->dimension != coordinates.size() ) cout << "DT Grid Push Operation - Error in dimension!" << endl;

    int coord = coordinates.back();
    coordinates.pop_back();

    // grid point lies in a new p-column
    if ( (this->projND != NULL) && ((this->projND->nCoord.size() == 0) || (this->projND->nCoord.back() != coordinates.back())) ) {
        // index pair into the values array and in the nCoord Array
        indexPair p;
        p.iValue = this->value.size();
        p.iCoord = this->nCoord.size();

        this->projND->push( coordinates, p );           // rekursive call of push in the next lower dt grid with the remaining coordinates and the index pair

        this->nCoord.push_back( coord );
        this->nCoord.push_back( coord );
        this->acc.push_back( this->value.size() );
    }
    // grid point is the first in a connected component
    else if ( (this->nCoord.size() == 0) ||(this->nCoord.back() != (coord -1)) ) {
        this->nCoord.push_back( coord );
        this->nCoord.push_back( coord );
        this->acc.push_back( this->value.size() );
    }
    // grid point is the last in a connected component
    else {
        this->nCoord.pop_back();
        this->nCoord.push_back( coord );
    }

    this->value.push_back( value );
}

/*!
    \fn CDtGrid::push( vector<int> coordinates, indexPair value )
 */
void CDtGrid::push( vector<int> coordinates, indexPair value )
{
    if ( this->dimension != coordinates.size() ) cout << "DT Grid Push Operation - Error in dimension!" << endl;

    int coord = coordinates.back();
    coordinates.pop_back();

    // grid point lies in a new p-column
    if ( (this->projND != NULL) && ((this->projND->nCoord.size() == 0) || (this->projND->nCoord.back() != coordinates.back())) ) {
        // index pair into the values array and in the nCoord Array
        indexPair p;
        p.iValue = this->valueIndexPair.size();
        p.iCoord = this->nCoord.size();

        this->projND->push( coordinates, p );           // rekursive call of push in the next lower dt grid with the remaining coordinates and the index pair

        this->nCoord.push_back( coord );
        this->nCoord.push_back( coord );
        this->acc.push_back( this->valueIndexPair.size() );
    }
    // grid point is the first in a connected component
    else if ( (this->nCoord.size() == 0) ||(this->nCoord.back() != (coord -1)) ) {
        this->nCoord.push_back( coord );
        this->nCoord.push_back( coord );
        this->acc.push_back( this->valueIndexPair.size() );
    }
    // grid point is the last in a connected component
    else {
        this->nCoord.pop_back();
        this->nCoord.push_back( coord );
    }

    this->valueIndexPair.push_back( value );
}

/*!
    \fn CDtGrid::create_hrle_from_voxel_object( CVoxelObject3D* vo )
 */
// convert an CVoxelObject3D into a Dynamic Tubular Grid
bool CDtGrid::create_dtgrid_from_voxel_object( CVoxelObject3D* vo ) {

    vector<int> coordinates;

    if ( this->dimension == 3 ) {

        // set minimum and maximum coordinates in the world coordinate system for drawing
        this->minRange = vo->m_MinRange;
        this->maxRange = vo->m_MaxRange;
//         this->projND->minRange = vo->m_MinRange.getY();
//         this->projND->maxRange = vo->m_MaxRange.getY();
//         this->projND->projND->minRange = vo->m_MinRange.getX();
//         this->projND->projND->maxRange = vo->m_MaxRange.getX();

        // extends from the bounding box for drawing
        this->minBBox = 0;
        this->maxBBox = vo->m_ScopeZ;
        this->projND->minBBox = 0;
        this->projND->maxBBox = vo->m_ScopeY;
        this->projND->projND->minBBox = 0;
        this->projND->projND->maxBBox = vo->m_ScopeX;

        for ( int x = 0; x < vo->m_ScopeX; x++ ) {
            for ( int y = 0; y < vo->m_ScopeY; y++ ) {
                for ( int z = 0; z < vo->m_ScopeZ; z++ ) {
                    coordinates.clear();

                    if ( vo->m_CubeArray(x,y,z).flag == CT_BORDER ) {
// cout << "Pushing Grid Point ("<< x << "," << y << "," << z <<")" << endl;
                        coordinates.push_back( x );
                        coordinates.push_back( y );
                        coordinates.push_back( z );

                        this->push( coordinates, 1.0 );
                    }
                }
            }
        }

    }
    else {
        cerr << "3D voxel object given so this must be a 3D DT Grid!!" << endl;
        return false;
    }

    return true;

}

/*!
    \fn CDtGrid::accessBinSearch( int coord, int kMin, int kMax );
 */
// "binary search" for the random access method
int CDtGrid::accessBinSearch( int coord, int kMin, int kMax )
{
    // p column has only one connected component
    if (kMin == kMax) return kMin;

    // otherwise
    int mid = (kMax - kMin) / 2 + kMin;

    if ( mid == 0 ) return mid;

    if ( ((mid % 2) == 0) && (this->nCoord.at(mid) <= coord) && (((mid+1) < this->nCoord.size()) && (this->nCoord.at(mid+1) > coord)) ) 
        return mid;
    else if ( this->nCoord.at(mid) > coord ) {
        if ((mid % 2) == 0) return this->accessBinSearch( coord, kMin, mid-2 );
        else return this->accessBinSearch( coord, kMin, mid-1 );
    }
    else {
        if ((mid % 2) == 0) return this->accessBinSearch( coord, mid+2, kMax );
        else return this->accessBinSearch( coord, mid+1, kMax );
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////
//                      Testing                                                             //
//////////////////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CHierarchicalRLE::draw()
 */
// draw the level set as cubes - only for 3d level sets implemented
void CDtGrid::drawCubes()
{
    // for comnputing the average access time of the random access operator
    long            numGridPoints = 0, accessTime = 0, totalAccessTime = 0, averageAccessTime = 0;
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)
    triple gridpoint;

    if (this->dimension == 3) {

        int minX = this->projND->projND->minBBox;
        int maxX = this->projND->projND->maxBBox;
        int minY = this->projND->minBBox;
        int maxY = this->projND->maxBBox;
        int minZ = this->minBBox;
        int maxZ = this->maxBBox;

        float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
        float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
        float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

        for (int z = minZ; z < maxZ; z++) {
            for (int y = minY; y < maxY; y++) {
                for (int x = minX; x < maxX; x++) {

                    gettimeofday( &tp1, NULL );
                    gridpoint = this->randomAccess( x, y, z );
                    gettimeofday( &tp2, NULL );

                    accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);
                    totalAccessTime += accessTime;
                    numGridPoints++;

                    if (gridpoint.inside == true) {

                        // draw cubes and cube edges
                        // -------------------------

                        CVertex3D point1( this->minRange.getX() + x * CubeW,
                                          this->minRange.getY() + y * CubeH,
                                          this->minRange.getZ() + z * CubeD );
                        CVertex3D point2( this->minRange.getX() + (x+1) * CubeW,
                                          this->minRange.getY() + (y+1) * CubeH,
                                          this->minRange.getZ() + (z+1) * CubeD);

                        this->drawACube(&point1, &point2, gridpoint.value);

                    }
                }
            }
        }

    }

    averageAccessTime = totalAccessTime / numGridPoints;

    cout << "Average access time for the Random Access Operator while drawing the whole object was " << averageAccessTime << " micro seconds." << endl;

    this->accessTimeRandomAccess = (double) accessTime;
    this->accessAverageTimeRandomAccess = (double) totalAccessTime / (double) numGridPoints;
}

/*!
    \fn CHierarchicalRLE::draw()
 */
// draw the level set as cubes (using the sequential access) - only for 3d level sets implemented
void CDtGrid::drawCubesUsingSequentialAccess()
{

    long            accessTime = 0;     // the access time for the function
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)

    if (this->dimension == 3) {

        int minX = this->projND->projND->minBBox;
        int maxX = this->projND->projND->maxBBox;
        int minY = this->projND->minBBox;
        int maxY = this->projND->maxBBox;
        int minZ = this->minBBox;
        int maxZ = this->maxBBox;

        float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
        float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
        float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

        // get all defined voxels
        gettimeofday( &tp1, NULL );
        vector<coordinate_value> definedCoordinates = this->sequentialAccess();
        gettimeofday( &tp2, NULL );

        for (int i = 0; i < definedCoordinates.size(); i++) {

            vector<int> coordinate = definedCoordinates[i].coordinates;
            double value = definedCoordinates[i].value;

            CVertex3D point1( this->minRange.getX() + coordinate[2] * CubeW,
                              this->minRange.getY() + coordinate[1] * CubeH,
                              this->minRange.getZ() + coordinate[0] * CubeD );
            CVertex3D point2( this->minRange.getX() + (coordinate[2]+1) * CubeW,
                              this->minRange.getY() + (coordinate[1]+1) * CubeH,
                              this->minRange.getZ() + (coordinate[0]+1) * CubeD);

            this->drawACube(&point1, &point2, value);

        }

        accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);

    }

    cout << "Access time for the Sequential Access Operator was " << accessTime << " micro seconds." << endl;
    this->accessTimeSequentialAccess = (double) accessTime;
}

void CDtGrid::drawACube(CVertex3D *p1, CVertex3D *p2, double colorValue) {

    // Fl�hen zeichnen
    // ----------------
        // Farbe festlegen
        glColor3f(0.0, 1.0, 0.5 * colorValue);

        // Offset fr den Cube (bessere Darstellung der Kanten)
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(1.f, 1.f);

            glBegin(GL_POLYGON);
                glNormal3f(  -1.0,  0.0,  0.0);
                glVertex3f(  p1->getX(), p1->getY(), p1->getZ());
                glVertex3f(  p1->getX(), p1->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d( 1.0,  0.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  -1.0,  0.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0, 1.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, -1.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, 1.0);
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

        glDisable(GL_POLYGON_OFFSET_FILL);


// Kanten zeichnen
// ---------------

        glColor3d(0.2, 0.2, 0.2);

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();


    return;
}

void CDtGrid::printDataStructure() {

    cout << "Ausgabe der Datenstruktur, Dimension: " << this->dimension << endl;
    int i;
//     cout << "Min: " << this->min << endl;
//     cout << "Max: " << this->max << endl;
    cout << "acc (" << this->acc.size() << "): ";
    for (i = 0; i < this->acc.size(); i++) cout << this->acc.at(i) << " | ";
    cout << endl;
    cout << "nCoord (" << this->nCoord.size() << "): ";
    for (i = 0; i < this->nCoord.size(); i++) cout << this->nCoord.at(i) << " | ";
    cout << endl;
    if ( this->parent == NULL ) {
        cout << "Values (" << this->value.size() << "): ";
        for (i = 0; i < this->value.size(); i++) cout << this->value.at(i) << " | ";
        cout << endl;
    }
    else {
        cout << "Values (" << this->valueIndexPair.size() << "): ";
        indexPair iPair;
        for (i = 0; i < this->valueIndexPair.size(); i++) {
            iPair = this->valueIndexPair.at(i);
            cout << "(" << iPair.iValue << "," << iPair.iCoord << ") | ";
        }
        cout << endl;
    }

    if ( this->projND != NULL ) this->projND->printDataStructure();


}

/*!
    \fn CDtGrid::getMemorySize()
 */
// returns the size of the data structure in bytes
double CDtGrid::getMemorySize()
{
    double returnValue;

    returnValue += this->acc.size() * sizeof(unsigned int);
    returnValue += this->nCoord.size() * sizeof(int);
    returnValue += this->value.size() * sizeof(double);
    returnValue += this->valueIndexPair.size() * sizeof(indexPair);

    // add the used memory of the lower dimensions
    if ( this->projND != NULL) returnValue += this->projND->getMemorySize();

    return returnValue;
}
