/***************************************************************************
 *   Copyright (C) 2006 by Holger Gerth                                    *
 *   holger.gerth@informatik.tu-chemnitz.de                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "hierarchicalrle.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//      Public functions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// constructor
CHierarchicalRLE::CHierarchicalRLE()
{

    this->dimension = 1;
    this->lower_hrle = NULL;

    // initialize min and max
    this->min = 0;
    this->max = 0;

    // for testing only
    this->accessTimeRandomAccess = 0.0;
    this->accessAverageTimeRandomAccess = 0.0;
    this->accessTimeSequentialAccess = 0.0;

}

// constructor
CHierarchicalRLE::CHierarchicalRLE( unsigned int dim )
{

    this->dimension = dim;

    if (dim > 1 ) {
        this->lower_hrle = new CHierarchicalRLE( dim -1 );
    }
    else {
        this->lower_hrle = NULL;
    }

    // initialize min and max
    this->min = 0;
    this->max = 0;

    // for testing only
    this->accessTimeRandomAccess = 0.0;
    this->accessAverageTimeRandomAccess = 0.0;
    this->accessTimeSequentialAccess = 0.0;

}

// constructor
CHierarchicalRLE::CHierarchicalRLE( unsigned int dim, CVoxelObject3D* vo )
{

    if (dim > 1 ) {
        this->dimension = dim;
        this->lower_hrle = new CHierarchicalRLE( dim -1 );
    }
    else {
        this->dimension = 1;
        this->lower_hrle = NULL;
    }

    // get the data from the voxel object
    this->create_hrle_from_voxel_object( vo );

    // for testing only
    this->accessTimeRandomAccess = 0.0;
    this->accessAverageTimeRandomAccess = 0.0;
    this->accessTimeSequentialAccess = 0.0;

}

// destructor
CHierarchicalRLE::~CHierarchicalRLE()
{
//     if ( this->lower_hrle != NULL ) delete this->lower_hrle;
}

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////

/*!
    \fn CHierarchicalRLE::convert_to_hrle( CVoxelObject3D* vo )
 */
// converts a voxel object into a hrle level set
bool CHierarchicalRLE::convert_to_hrle( CVoxelObject3D* vo ) {

    if (this->dimension != 3) {
        cerr << "HRLE Object has false dimension - need to be 3!" << endl;
        return false;
    }
    else {

        CVoxelObject3D* tmp_vo = new CVoxelObject3D();
        bool returnValue;

        tmp_vo->Copy( vo );

        returnValue = this->create_hrle_from_voxel_object( tmp_vo );

        delete tmp_vo;

        return returnValue;
    }

}


/*!
    \fn CHierarchicalRLE::randomAccess( vector<int> *qv, int segmentIndex)
 */
// the random access operator to access data
run_code CHierarchicalRLE::randomAccess( vector<int> qv, int segmentIndex)
{
    int qi = qv.back();                                // split first coordinate
    qv.pop_back();                                     // delete the coordinate from the vector
    int segmentLength;
    int dataIndex;
    unsigned int firstRun;
    unsigned int firstRunBreak;
    int indexRun;                                       // index of the run in the run codes array
    run_code run;                                       // a run code and (if defined) the index in the defined data array
    int firstCoordinate = this->min;                    // minimum coordinate or last break before the determined run
    int lastCoordinate = this->max;                     // miximum coordinate or the break after the determined run

    // qi doesn't lie between the extends of this HRLE
    if ( (qi < this->min) || (qi >= this->max) ) {
        run.type = RT_POSITIVE;
        return run;
    }

    // get segment length - number of runs in this segment
    if (this->start_indices.at(segmentIndex) == this->start_indices.at(this->start_indices.size()-1))
        segmentLength = this->run_types.size() - this->start_indices.at(segmentIndex);
    else
        segmentLength = this->start_indices.at(segmentIndex+1) - this->start_indices.at(segmentIndex);

    // get the index of the first run in this segment
    firstRun = this->start_indices.at(segmentIndex);

    // get the index of the first run break of this segment
    firstRunBreak = this->start_indices.at(segmentIndex) - segmentIndex;

    // get the index of the run containing qi
    indexRun = this->accessBinSearch( qi, firstRun, firstRunBreak, segmentLength, &firstCoordinate, &lastCoordinate );

    // get the run code
    run = this->run_types.at(indexRun);

    if (run.type == RT_DEFINED) {

        // defined data index = run code + offset of qi
        dataIndex = run.index + qi - firstCoordinate;

        // recursive traversal to the next lower hrle block
        if (this->lower_hrle != NULL) { 
            return this->lower_hrle->randomAccess( qv, dataIndex );
        }
        else run.index = dataIndex;                                     // return the direct index into the value array

    }

    return run;

}

run_code CHierarchicalRLE::randomAccess( int X, int Y, int Z, int segmentIndex) {

    // create a stl vector from the given coordinates
    vector<int>     qv;
    qv.push_back(X);
    qv.push_back(Y);
    qv.push_back(Z);

    return this->randomAccess( qv, segmentIndex);


}

/*!
    \fn CHierarchicalRLE::getDefinedValue( int dataIndex )
 */
// returns the value of the given defined data index
double CHierarchicalRLE::getDefinedValue( int dataIndex )
{
    if ( this->lower_hrle != NULL ) return this->lower_hrle->getDefinedValue( dataIndex );
    else return this->defined_values.at( dataIndex );
}

/*!
    \fn CHierarchicalRLE::neighborAccess( vector<int> *qv, int segmentIndex)
 */
// the random access operator to access data
vector<neighbors> CHierarchicalRLE::neighborAccess( vector<int> qv, int axis, int segmentIndex)
{

    int qi = qv.back();                                 // split first coordinate
    int qi_left = qi-1;                                 // left neighbor of qi
    int qi_right = qi+1;                                // right neighbor of qi
    qv.pop_back();                                      // delete the coordinate from the vector
    int segmentLength;                                  // length of the current segment
    int dataIndex;                                      // defined data index
    int firstRun;                                       // first run of the current segment
    int firstRunBreak;                                  // first break in this segment
    int indexRun;                                       // index of the run in the run codes array
    run_code run;                                       // a run code and (if defined) the index in the defined data array
    run_code run_left;                                  // run code of the left run
    run_code run_right;                                 // run code of the right run
    neighbors currentNeighbors;                         // neighbors at this axis
    currentNeighbors.leftNeighbor = false;
    currentNeighbors.rightNeighbor = false;
    vector<neighbors> returnValue;                      // run codes of the neighbors of given coordinate
    int firstCoordinate = this->min;                    // minimum coordinate or last break before the determined run
    int lastCoordinate = this->max;                     // miximum coordinate or the break after the determined run
    vector<int>     qv_left = qv;                       // temporary variable for left random access
    vector<int>     qv_right = qv;                      // temporary variable for right random access

    // get segment length - number of runs in this segment
    if (this->start_indices.at(segmentIndex) == this->start_indices.at(this->start_indices.size()-1))
        segmentLength = this->run_types.size() - this->start_indices.at(segmentIndex);
    else
        segmentLength = this->start_indices.at(segmentIndex+1) - this->start_indices.at(segmentIndex);

    // get the index of the first run in this segment
    firstRun = this->start_indices.at(segmentIndex);

    // get the index of the first run break of this segment
    firstRunBreak = this->start_indices.at(segmentIndex) - segmentIndex;

    // get the index of the run containing qi
    indexRun = this->accessBinSearch( qi, firstRun, firstRunBreak, segmentLength, &firstCoordinate, &lastCoordinate );

    // get the run code
    run = this->run_types.at(indexRun);

    if ( run.type != RT_DEFINED ) {
        cout << " This grid point is not defined so neighbors cannot be detected!" << endl;
        return returnValue;
    }

    // get the neighbors of this axis; 0 means all neighbors at each axis are wanted
    if ( ( axis == this->dimension) || ( axis == 0 ) ) {
        // get the neighbors
        // if the left neighbor is in the same run code like the given coordinate
        if ((firstCoordinate <= qi_left) && (qi_left < lastCoordinate)) {

            if (run.type == RT_DEFINED) {
                dataIndex = run.index + qi_left - firstCoordinate;
                currentNeighbors.leftNeighbor = true;
                // if this is the lowest dimension get the value, else random access to the neighbor segment in the next lower HRLE
                if (this->lower_hrle == NULL)
                    currentNeighbors.leftValue = this->defined_values.at( dataIndex );
                else {
                    run_left = this->lower_hrle->randomAccess( qv_left, dataIndex );
                    if ( run_left.type == RT_DEFINED ) {
                        currentNeighbors.leftNeighbor = true;
                        currentNeighbors.leftValue = this->getDefinedValue( run_left.index );
                    }
                    else currentNeighbors.leftNeighbor = false;
                }
            }
            else currentNeighbors.leftNeighbor = false;                 // the point has ne neighbor at qi-1

        }

        // if the left neighbor is at index-1
        else {
            if ( this->min <= qi_left) {

                // we need the minimal coordinate of this new run
                int left_break;
                run_left = this->getRunByCoordinate( qi_left, segmentIndex, &left_break );

                if (run_left.type == RT_DEFINED) {
                    dataIndex = run_left.index + qi_left - left_break;
                    currentNeighbors.leftNeighbor = true;
                    // if this is the lowest dimension get the value, else random access to the neighbor segment in the next lower HRLE
                    if (this->lower_hrle == NULL)
                        currentNeighbors.leftValue = this->defined_values.at( dataIndex );
                    else {
                        run_left = this->lower_hrle->randomAccess( qv_left, dataIndex );
                        if ( run_left.type == RT_DEFINED ) {
                            currentNeighbors.leftNeighbor = true;
                            currentNeighbors.leftValue = this->getDefinedValue( run_left.index );
                        }
                        else currentNeighbors.leftNeighbor = false;
                    }
                }
                else currentNeighbors.leftNeighbor = false;             // the point has ne neighbor at qi-1
            }
        }

        // if the left neighbor is in the same run like the given coordinate
        if ( (firstCoordinate <= qi_right) && (qi_right < lastCoordinate) ) {

            if (run.type == RT_DEFINED) {
                dataIndex = run.index + qi_right - firstCoordinate;
                currentNeighbors.rightNeighbor = true;
                // if this is the lowest dimension get the value, else random access to the neighbor segment in the next lower HRLE
                if (this->lower_hrle == NULL)
                    currentNeighbors.rightValue = this->defined_values.at( dataIndex );
                else {
                    run_right = this->lower_hrle->randomAccess( qv_right, dataIndex );
                    if ( run_right.type == RT_DEFINED ) {
                        currentNeighbors.rightNeighbor = true;
                        currentNeighbors.rightValue = this->getDefinedValue( run_right.index );
                    }
                    else currentNeighbors.rightNeighbor = false;
                }
            }
            else currentNeighbors.rightNeighbor = false;                 // the point has ne neighbor at qi+1

        }

        // if the left neighbor is at index-1
        else {
            if (qi_right < this->max) {
                // we need the minimal coordinate of this new run
                int right_break;
                run_right = this->getRunByCoordinate( qi_right, segmentIndex, &right_break );

                if (run_right.type == RT_DEFINED) {
                    dataIndex = run_right.index + qi_right - right_break;
                    currentNeighbors.rightNeighbor = true;
                    // if this is the lowest dimension get the value, else random access to the neighbor segment in the next lower HRLE
                    if (this->lower_hrle == NULL)
                        currentNeighbors.rightValue = this->defined_values.at( dataIndex );
                    else {
                        run_right = this->lower_hrle->randomAccess( qv_right, dataIndex );
                        if ( run_right.type == RT_DEFINED ) {
                            currentNeighbors.rightNeighbor = true;
                            currentNeighbors.rightValue = this->getDefinedValue( run_right.index );
                        }
                        else currentNeighbors.rightNeighbor = false;
                    }
                }
                else currentNeighbors.rightNeighbor = false;                 // the point has ne neighbor at qi+1
            }
        }

        // neighbors found in this dimension, so add them to the return value
        if ( (currentNeighbors.rightNeighbor == true) || (currentNeighbors.leftNeighbor == true) ) {
            currentNeighbors.dimension = this->dimension;
            returnValue.push_back( currentNeighbors );

        }
    }

    // this dimension is net the demanded or we want all neighbors
    if ( (this->dimension > axis) || (axis == 0) ) {
        // get the neighbors in the next lower dimension
        if ( (run.type == RT_DEFINED) && (this->lower_hrle != NULL) ) {
            dataIndex = run.index + qi - firstCoordinate;
            // neighbor access into the next lower dimension
            vector<neighbors> tmp = this->lower_hrle->neighborAccess( qv, axis, dataIndex );
            // add the neighbors of the lower dimensions to the return value
            for (int i = 0; i < tmp.size(); i++) {
                returnValue.push_back( tmp[i] );
                cout << "i: " << i << endl;
            }

        }
    }

    return returnValue;

}

vector<neighbors> CHierarchicalRLE::neighborAccess( int X, int Y, int Z, int axis, int segmentIndex) {

    // create a stl vector from the given coordinates
    vector<int>     qv;
    qv.push_back(X);
    qv.push_back(Y);
    qv.push_back(Z);

    return this->neighborAccess( qv, axis, segmentIndex);


}


/*!
    \fn CHierarchicalRLE::sequentialAccess()
 */
// sequential access to the H-RLE level set
vector<coordinate_value> CHierarchicalRLE::sequentialAccess()
{
    vector<int> parentVector;
    return this->sequentialAccess( parentVector );
}

/*!
    \fn CHierarchicalRLE::sequentialAccess( vector<int> parentCoord, int segmentIndex )
 */
// sequential access to the H-RLE level set
vector<coordinate_value> CHierarchicalRLE::sequentialAccess( vector<int> parentCoord, int segmentIndex )
{
    int segmentLength;
    int firstRun = this->start_indices.at(segmentIndex);
    run_code run;
    int startCoordinate;                                        // the first coordinate of the determined run
    int runLength;                                              // the count of coordinates in the run
    vector<coordinate_value> returnValue;

    // get segment length - number of runs in this segment
    if (this->start_indices.at(segmentIndex) == this->start_indices.at(this->start_indices.size()-1))
        segmentLength = this->run_types.size() - this->start_indices.at(segmentIndex);
    else
        segmentLength = this->start_indices.at(segmentIndex+1) - this->start_indices.at(segmentIndex);

    // traverse the runs of this segment
    for (int i = firstRun; i < ( firstRun + segmentLength ); i++) {
        run = this->run_types.at(i);

        // only defined runs are considered
        if ( run.type == RT_DEFINED ) {

            // get the first coordinate of the run
            if ( i == firstRun ) {
                startCoordinate = this->min;
            }
            else {
                startCoordinate = this->run_breaks.at( i - segmentIndex -1 );
            }

            // length of the run
            if (i == ( firstRun + segmentLength -1 )) {
                runLength = this->max - startCoordinate;
            }
            else {
                runLength = this->run_breaks.at( i - segmentIndex ) - startCoordinate;
            }

            // iterate trough all coordinates of the run
            for (int j = startCoordinate; j < (startCoordinate + runLength); j++) {
                int defDataIndex = run.index + j - startCoordinate;             // compute the defined data index
                vector<int> currentCoordinates = parentCoord;                   // concatenate the parent coordinates with current coordinate
                currentCoordinates.push_back( j );

                // if this is the lowest dimension, return the current coordinates and the coresponding value
                if (this->lower_hrle == NULL) {
                    coordinate_value newCoord;
                    newCoord.coordinates = currentCoordinates;
                    newCoord.index = defDataIndex;
                    returnValue.push_back( newCoord );
                }
                // recursive call to the next lower H_RLE block
                else {
                    vector<coordinate_value> tmp;
                    tmp = this->lower_hrle->sequentialAccess( currentCoordinates, defDataIndex );
                    for (int k = 0; k < tmp.size(); k++)
                        returnValue.push_back(tmp.at(k));
                }

            }

        }
    }

    return returnValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//      Private functions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CHierarchicalRLE::create_hrle_from_voxel_object( CVoxelObject3D* vo )
 */
// convert an CVoxelObject3D into the HRLE structure
bool CHierarchicalRLE::create_hrle_from_voxel_object( CVoxelObject3D* vo ) {

    /// @TODO Funktionsaufbau vereinfachen, zu unbersichtlich

    int segments = 0;
    int numRunCodes;

    this->minRange = vo->m_MinRange;
    this->maxRange = vo->m_MaxRange;

    // create x-RLE Block
    if (this->dimension == 1) {

        this->min = 0;
        this->max = vo->m_ScopeX;

        for (int z = 0; z < vo->m_ScopeZ; z++) {
            for (int y = 0; y < vo->m_ScopeY; y++) {
                numRunCodes = 0;
                segments++;

                for (int x = 0; x < vo->m_ScopeX; x++) {

                    // handle blocks which are outer
                    if (vo->m_CubeArray(x,y,z).flag == CT_EMPTY) {
                        if (numRunCodes == 0) {
                            run_code newRunCode;
                            newRunCode.type = RT_POSITIVE;
                            this->run_types.push_back(newRunCode);
                            numRunCodes++;
                        }
                        else if (this->run_types[this->run_types.size()-1].type == RT_DEFINED) {
                            this->run_breaks.push_back(x);
                            run_code newRunCode;
                            newRunCode.type = RT_POSITIVE;
                            this->run_types.push_back(newRunCode);
                            numRunCodes++;
                        }
                    }

                    // handle the border blocks
                    else if (vo->m_CubeArray(x,y,z).flag == CT_BORDER) {
                        if (numRunCodes == 0) {
                            run_code newRunCode;
                            newRunCode.type = RT_DEFINED;
                            newRunCode.index = this->defined_values.size();
                            this->run_types.push_back(newRunCode);
                            numRunCodes++;
                        }
                        else if (this->run_types[this->run_types.size()-1].type != RT_DEFINED) {
                            this->run_breaks.push_back(x);
                            run_code newRunCode;
                            newRunCode.type = RT_DEFINED;
                            newRunCode.index = this->defined_values.size();
                            this->run_types.push_back(newRunCode);
                            numRunCodes++;
                        }

                        this->defined_values.push_back(1.0);
                    }

                    // handle blocks which are inner
                    else if (vo->m_CubeArray(x,y,z).flag == CT_INNER) {
                        if (this->run_types[this->run_types.size()-1].type == RT_DEFINED) {
                            run_code newRunCode;
                            newRunCode.type = RT_NEGATIVE;
                            this->run_types.push_back(newRunCode);
                            this->run_breaks.push_back(x);
                            numRunCodes++;
                        }
                    }
                }

                // if no defined runs are in this segment
                if ((numRunCodes == 1) && (this->run_types[this->run_types.size()-1].type != RT_DEFINED)) {
                    // remove last run -> will be saved in the next higher dimension
                    this->run_types.pop_back();
                }
                else {
                    // add offset of this segment to the start_indices block
                    this->start_indices.push_back( this->run_types.size() - numRunCodes );

                    // create a 2D projection of the voxel set
                    vo->m_CubeArray(0,y,z).flag = CT_BORDER;
                }
            }
        }

    }

    // start encoding along Y axis
    else if (this->dimension == 2) {

        this->lower_hrle->create_hrle_from_voxel_object( vo );

        this->min = 0;
        this->max = vo->m_ScopeY;

        int segmentIndex = 0;

        for (int z = 0; z < vo->m_ScopeZ; z++) {
            numRunCodes = 0;
            segments++;
            for (int y = 0; y < vo->m_ScopeY; y++) {
                // handle blocks which are outer
                if (vo->m_CubeArray(0,y,z).flag == CT_EMPTY) {
                    if (numRunCodes == 0) {
                        run_code newRunCode;
                        newRunCode.type = RT_POSITIVE;
                        this->run_types.push_back(newRunCode);
                        numRunCodes++;
                    }
                    else if (this->run_types[this->run_types.size()-1].type == RT_DEFINED) {
                        this->run_breaks.push_back(y);
                        run_code newRunCode;
                        newRunCode.type = RT_POSITIVE;
                        this->run_types.push_back(newRunCode);
                        numRunCodes++;
                    }
                }

                // handle the border blocks
                else if (vo->m_CubeArray(0,y,z).flag == CT_BORDER) {
                    if (numRunCodes == 0) {
                        run_code newRunCode;
                        newRunCode.type = RT_DEFINED;
                        newRunCode.index = segmentIndex;
                        this->run_types.push_back(newRunCode);
                        numRunCodes++;
                    }
                    else if (this->run_types[this->run_types.size()-1].type != RT_DEFINED) {
                        this->run_breaks.push_back(y);
                        run_code newRunCode;
                        newRunCode.type = RT_DEFINED;
                        newRunCode.index = segmentIndex;
                        this->run_types.push_back(newRunCode);
                        numRunCodes++;
                    }

                    segmentIndex++;

                }

                // handle blocks which are inner
                else if (vo->m_CubeArray(0,y,z).flag == CT_INNER) {
                    if (this->run_types[this->run_types.size()-1].type == RT_DEFINED) {
                        run_code newRunCode;
                        newRunCode.type = RT_NEGATIVE;
                        this->run_types.push_back(newRunCode);
                        this->run_breaks.push_back(y);
                        numRunCodes++;
                    }
                }
            }

            // if no defined runs are in this segment
            if ((numRunCodes == 1) && (this->run_types[this->run_types.size()-1].type != RT_DEFINED)) {
                // remove last run -> will be saved in the next higher dimension
                this->run_types.pop_back();
            }
            else {
                // add offset of this segment to the start_indices block
                this->start_indices.push_back( this->run_types.size() - numRunCodes );

                // create a 2D projection of the voxel set
                vo->m_CubeArray(0,0,z).flag = CT_BORDER;

            }
        }

    }

    // start encoding along Z axis
    else if (this->dimension == 3) {

        this->lower_hrle->create_hrle_from_voxel_object( vo );

        this->min = 0;
        this->max = vo->m_ScopeZ;
        cout << "ScopeZ: " << vo->m_ScopeZ << endl;

        int segmentIndex = 0;
        numRunCodes = 0;

        for (int z = 0; z < vo->m_ScopeZ; z++) {
            // handle blocks which are outer
            if (vo->m_CubeArray(0,0,z).flag == CT_EMPTY) {
                if (numRunCodes == 0) {
                    run_code newRunCode;
                    newRunCode.type = RT_POSITIVE;
                    this->run_types.push_back(newRunCode);
                    numRunCodes++;
                }
                else if (this->run_types[this->run_types.size()-1].type == RT_DEFINED) {
                    this->run_breaks.push_back(z);
                    run_code newRunCode;
                    newRunCode.type = RT_POSITIVE;
                    this->run_types.push_back(newRunCode);
                    numRunCodes++;
                }
            }

            // handle the border blocks
            else if (vo->m_CubeArray(0,0,z).flag == CT_BORDER) {
                if (numRunCodes == 0) {
                    run_code newRunCode;
                    newRunCode.type = RT_DEFINED;
                    newRunCode.index = segmentIndex;
                    this->run_types.push_back(newRunCode);
                    numRunCodes++;
                }
                else if (this->run_types[this->run_types.size()-1].type != RT_DEFINED) {
                    this->run_breaks.push_back(z);
                    run_code newRunCode;
                    newRunCode.type = RT_DEFINED;
                    newRunCode.index = segmentIndex;
                    this->run_types.push_back(newRunCode);
                    numRunCodes++;
                }

                segmentIndex++;

            }

            // handle blocks which are inner
            else if (vo->m_CubeArray(0,0,z).flag == CT_INNER) {
                if (this->run_types[this->run_types.size()-1].type == RT_DEFINED) {
                    run_code newRunCode;
                    newRunCode.type = RT_NEGATIVE;
                    this->run_types.push_back(newRunCode);
                    this->run_breaks.push_back(z);
                    numRunCodes++;
                }
            }
        }

        this->start_indices.push_back( 0 );

    }
    else return false;

// cout << this->max << endl;

    return true;

}

/*!
    \fn CHierarchicalRLE::rebuild ( double distance )
 */
// rebuild the Hierarchical RLE Level Set
CHierarchicalRLE* CHierarchicalRLE::rebuild ( double distance ) {

    // remove grid points which are to far away from the interface
    CHierarchicalRLE* tmp = this->removeFarGridPoints( distance );

    // dilate the remaining grid points
    return tmp->dilation( 1 );

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                      Private Classes                                                                                 //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CHierarchicalRLE::removeFarGridPoints ( double distance, int segmentIndex = 0  )
 */
// remove grid points which are to far away from the interface (defined value is greater distance or lower -distance)
CHierarchicalRLE* CHierarchicalRLE::removeFarGridPoints ( double distance, int segmentIndex ) {

    CHierarchicalRLE* returnValue = new CHierarchicalRLE();
    CHierarchicalRLE* tmp = NULL;
    run_code tmpRun, lastRun;
    int segmentStart, segmentEnd, segmentLength, startCoordinate, endCoordinate;
    int numDefinedRuns = 0;
    double definedData;

    if ( distance < 0.0 ) {
        cerr << "Error: negative distance in removeFarGridPoints!" << endl;
        return NULL;
    }


    // get start and end indices of this segment
    segmentStart = this->start_indices.at(segmentIndex);
    if ((segmentIndex+1) >= this->start_indices.size())
        segmentEnd = this->run_types.size();
    else
        segmentEnd = this->start_indices.at(segmentIndex+1);

    // length of this segment
    segmentLength = segmentEnd - segmentStart;

    // start and end coordinate of the first run (end coordinate don't belong to the run!)
    startCoordinate = this->min;                                                                   // break before the run
    if (segmentLength < 2)  endCoordinate = this->max;
    else endCoordinate = this->run_breaks.at(this->start_indices.at(segmentIndex) - segmentIndex); // break after the run

    // max & min
    returnValue->min = this->min;
    returnValue->max = this->max;

    // 1D so wo only need one start index
    returnValue->start_indices.push_back( 0 );

    for ( int i = segmentStart; i < segmentEnd; i++ ) {
        tmpRun = this->run_types.at(i);

        // we must check the defined values of each defined block if it is further away from the interface than distance
        if ( tmpRun.type == RT_DEFINED ) {
            for ( int qi = startCoordinate; qi < endCoordinate; qi ++ ) {

                // compute defined index
                int definedIndex = tmpRun.index + qi - startCoordinate;

                // higher dimension
                if ( this->lower_hrle != NULL ) {

                    // recursive call in the next lower dimension
                    tmp = this->lower_hrle->removeFarGridPoints( distance, definedIndex );

                    // there exists lower segements
                    if ( tmp != NULL ) {

                        if ( ( returnValue->run_types.size() > 0 ) && ( lastRun.type == RT_DEFINED ) ) {
                            if ( returnValue->run_breaks.size() > 0) returnValue->run_breaks.pop_back();
                            returnValue->run_breaks.push_back(qi+1);
                        }
                        else {
                            run_code newRun;
                            newRun.type = RT_DEFINED;
                            if ( returnValue->lower_hrle != NULL ) newRun.index = returnValue->lower_hrle->start_indices.size();
                            else newRun.index = 0;
                            returnValue->run_types.push_back( newRun );
                            returnValue->run_breaks.push_back( qi +1 );
                            // set last run
                            lastRun = newRun;
                            numDefinedRuns++;
                        }

                        if (returnValue->lower_hrle == NULL) returnValue->lower_hrle = tmp;
                        else returnValue->lower_hrle->addToHRLE( tmp );

                    }
                    // this coordinate is now non defined
                    else {
                        // last run was outside or inside
                        if ( (returnValue->run_types.size() > 0) && (lastRun.type != RT_DEFINED) ) {
                            if ( returnValue->run_breaks.size() > 0) returnValue->run_breaks.pop_back();
                            returnValue->run_breaks.push_back(qi+1);
                        }
                        // last run was defined; replace the defined run with an inside run
                        else {
                            run_code newRun;
                            newRun.type = RT_NEGATIVE;
                            returnValue->run_types.push_back( newRun );
                            returnValue->run_breaks.push_back( qi +1 );
                            // set last run
                            lastRun = newRun;
                        }
                    }
                }
                // lowest dimension
                else {
                    // get defined data
                    definedData = this->defined_values.at( definedIndex );

                    // skip grid points which are further away from the interface than distance
                    if ( (definedData <= distance) && (definedData >= -distance) ) {
                        if ( ( returnValue->run_types.size() > 0 ) && ( lastRun.type == RT_DEFINED ) ) {
                            if ( returnValue->run_breaks.size() > 0) returnValue->run_breaks.pop_back();
                            returnValue->run_breaks.push_back(qi+1);
                        }
                        else {
                            run_code newRun;
                            newRun.type = RT_DEFINED;
                            newRun.index = returnValue->defined_values.size();
                            returnValue->run_types.push_back( newRun );
                            returnValue->run_breaks.push_back( qi +1 );
                            // set last run
                            lastRun = newRun;
                            numDefinedRuns++;
                        }
                        returnValue->defined_values.push_back( definedData );

                    }
                    // this coordinate is now non defined
                    else {
                        // last run was outside or inside
                        if ( (returnValue->run_types.size() > 0) && (lastRun.type != RT_DEFINED) ) {
                            if ( returnValue->run_breaks.size() > 0) returnValue->run_breaks.pop_back();
                            returnValue->run_breaks.push_back(qi+1);
                        }
                        // last run was defined; replace the defined run with an inside run
                        else {
                            run_code newRun;
                            newRun.type = RT_NEGATIVE;
                            returnValue->run_types.push_back( newRun );
                            returnValue->run_breaks.push_back( qi +1 );
                            // set last run
                            lastRun = newRun;
                        }
                    }
                }

            }
        }
        // add non defined runs
        else {

            // if the last run is the same, pop the last break
            if ( (returnValue->run_types.size() > 0) && (lastRun.type == tmpRun.type) )
                returnValue->run_breaks.pop_back();
            // last run was negative, this is positive
            else if ( (returnValue->run_types.size() > 0) && (lastRun.type == RT_NEGATIVE) ) {
                returnValue->run_types.pop_back();
                returnValue->run_breaks.pop_back();
                tmpRun.type = RT_POSITIVE;
                returnValue->run_types.push_back( tmpRun );
            }
            // last run was defined, push the current run
            else
                returnValue->run_types.push_back( tmpRun );

            // add break
            returnValue->run_breaks.push_back( endCoordinate );

        }

        // update start- and endcoordinate
        startCoordinate = endCoordinate;
        if ((i+2) < segmentEnd)  endCoordinate = this->run_breaks.at((i+1) - segmentIndex);
        else endCoordinate = this->max;

        // get last run
        lastRun = returnValue->run_types.back();
    }

    if ( returnValue->run_breaks.back() == returnValue->max ) returnValue->run_breaks.pop_back();

    if ( numDefinedRuns > 0 ) {
        returnValue->dimension = this->dimension;
        return returnValue;
    }
    else {
        if (returnValue != NULL) delete returnValue;
        return NULL;
    }

}


/*!
    \fn CHierarchicalRLE::accessBinSearch( int qi, unsigned int indexFirstRun, unsigned int indexFirstBreak, int segmentLength, int *min, int *max)
 */
// Binary Search within the run breaks array . returns index in the start_indices array
int CHierarchicalRLE::accessBinSearch( int qi, unsigned int indexFirstRun, unsigned int indexFirstBreak, int segmentLength, int *min, int *max )
{

    if ( segmentLength == 1 ) {
        return indexFirstRun;
    }
    if ( segmentLength == 2 ) {
        if ( qi < this->run_breaks.at(indexFirstBreak)) {
            *max = this->run_breaks.at(indexFirstBreak);
            return indexFirstRun;
        }
        else {
            *min = this->run_breaks.at(indexFirstBreak);
            return (indexFirstRun+1);
        }
    }
    else {

        // get middle element of the run breaks
        int mid = indexFirstBreak + ((segmentLength - 1) / 2);

        // recursive search in the left array
        if ( qi < this->run_breaks.at(mid)) {
           *max = this->run_breaks.at(mid);
           return this->accessBinSearch( qi, indexFirstRun, indexFirstBreak, (segmentLength+1) / 2, min, max);
        }
        // recursive search in the right array
        else {
           *min = this->run_breaks.at(mid);
           return this->accessBinSearch( qi, indexFirstRun + ((segmentLength+1) / 2), indexFirstBreak + ((segmentLength+1) / 2),
                                         segmentLength / 2, min, max);
        }
    }
}

/*!
    \fn CHierarchicalRLE::dilation( int pointCount )
 */
// dilation algorithm for the rebuilding of the H-RLE Level Set
CHierarchicalRLE* CHierarchicalRLE::dilation( int pointCount )
{
    if (pointCount < 1) cout << "Pointcount must be at least 1 for dilation!!!!" << endl;

    // dilate the highest RLE block which consists only of one segment
    CHierarchicalRLE* dilatedHRLE = this->dilateSegment( 0, pointCount );

// cout << "------------------------- Original H-RLE -------------------------" << endl;
// this->printDataStructure();
// cout << "------------------------- -------------- -------------------------" << endl;

    // next the dilation of the whole data structure is started
    dilatedHRLE = this->dilation( dilatedHRLE, pointCount );

// cout << "------------------------- Dilated H-RLE (without data)  -------------------------" << endl;
// dilatedHRLE->printDataStructure();
// cout << "------------------------- -------------- -------------------------" << endl;

    // update the values array - initialize new values and give existing grid points there original value
    dilatedHRLE->updateValueArray( this );

// cout << "------------------------- Dilated H-RLE  -------------------------" << endl;
// dilatedHRLE->printDataStructure();
// cout << "------------------------- -------------- -------------------------" << endl;

    return dilatedHRLE;

}


/*!
    \fn CHierarchicalRLE::dilation( CHierarchicalRLE dilatedHRLE, int pointCount )
 */
// dilation algorithm for the rebuilding of the H-RLE Level Set
CHierarchicalRLE* CHierarchicalRLE::dilation( CHierarchicalRLE* dilatedHRLE, int pointCount, int si )
{

    run_code                        run;
    int                             startCoordinate;
    int                             endCoordinate;
    run_code                        tmpRun;
    CHierarchicalRLE                *dilatedSubHRLE = NULL;         // the dilated lower RLE block
    CHierarchicalRLE                *tmpSubSegment = NULL;          // just one dilated subsegment
    int                             indexRunBreak = 0;
    vector<CHierarchicalRLE*>       tmpSubsegments;                 // all dilated subsegments in the range ym-n to ym+n

    // start- and endcoordinate of the first run
    startCoordinate = dilatedHRLE->min;
    if ( dilatedHRLE->run_types.size() < 2 ) endCoordinate = dilatedHRLE->max;
    else endCoordinate = dilatedHRLE->run_breaks.at(0);

    // now we dilate the next lower hrle-block (if it exists)
    if ( this->lower_hrle != NULL ) {

            // for all runs in this segment
            for ( int j = 0; j < dilatedHRLE->run_types.size(); j++) {
                run = dilatedHRLE->run_types.at(j);                                                // get run from the dilated hrle segment

                // dilate the subsegments if this run is defined
                if ( run.type == RT_DEFINED ) {                                                    // if this dilated run is defined

                    // update the defined run index
                    if ( dilatedSubHRLE == NULL )
                        dilatedHRLE->run_types[j].index = 0;
                    else 
                        dilatedHRLE->run_types[j].index = dilatedSubHRLE->start_indices.size();

                    // get dilated rle subsegments in the range ym-n to ym+n
                    for ( int ym = startCoordinate; ym < endCoordinate; ym++) {                    // traverse all coordinates of this defined run
                        tmpSubsegments.clear();                                                    // delete subsegments from the last coordinate

                        for ( int qi = (ym-pointCount); qi <= (ym+pointCount); qi++) {             // recursive dilation in the range ym-n to ym+n

                            if ( (qi >= dilatedHRLE->min) && (qi < dilatedHRLE->max) ) {           // check if the coordinates are between min and max
                                int startOfRun = this->min;
                                tmpRun = this->getRunByCoordinate( qi, si, &startOfRun );

                                if ( tmpRun.type == RT_DEFINED ) {                                 // check if the corresponding run is defined in the original RLE segment
                                    int definedIndex = tmpRun.index + qi - startOfRun;             // compute defined data index -> segment index of the corresponding lower hrle segment

                                    CHierarchicalRLE* tmpSegment = this->lower_hrle->dilateSegment( definedIndex, pointCount );  // dilate the subsegment

                                    // recursive call for the next lower dimension
                                    if ( this->lower_hrle->lower_hrle != NULL ) 
                                        tmpSegment = this->lower_hrle->dilation( tmpSegment, pointCount, definedIndex );

                                    // add this dilated subsegment for union
                                    tmpSubsegments.push_back( tmpSegment );
                                }
                            }
                        }

                        // union of the dilated subsegments
                        tmpSubSegment = this->hrleUnion( tmpSubsegments );

                        // add the new dilated subsegment to the dilated HRLE block
                        if ( dilatedSubHRLE == NULL )
                            dilatedSubHRLE = tmpSubSegment;
                        else 
                            dilatedSubHRLE->addToHRLE( tmpSubSegment );
                    }
                }

                // start- and endcoordinate of the next run
                startCoordinate = endCoordinate;
                if ( (j+1) >= (dilatedHRLE->run_types.size()-1) ) endCoordinate = dilatedHRLE->max;
                else endCoordinate = dilatedHRLE->run_breaks.at(++indexRunBreak);
            }

        // add the dilated lower RLE block to the dilated H-RLE data structure
        dilatedHRLE->lower_hrle = dilatedSubHRLE;

        // update dimension
        dilatedHRLE->dimension = this->dimension;
        return dilatedHRLE;
    }

    return NULL;
}

/*!
    \fn CHierarchicalRLE::dilateSegment( int si, int pointCount )
 */
// this is the 1D dilation algorithm
// only run types und run breaks are used! subsegments ans values are not considered
CHierarchicalRLE* CHierarchicalRLE::dilateSegment( int si, int pointCount )
{
    CHierarchicalRLE*               dilatedHRLE = new CHierarchicalRLE();       // the dilated hrle level set
    vector<int>                     dilatedCoordinates;                         // run break array after the dilationCArray3D
    run_code                        run;
    int                             segmentStart;                               // starting index of the segment
    int                             segmentEnd;                                 // starting index of the next segment
    int                             segmentLength;
    int                             startCoordinate;
    int                             endCoordinate;
    run_code                        tmpRun;
    int                             firstBreak;
    int                             j;

    // get start and end indices of this segment
    segmentStart = this->start_indices.at(si);
    if (segmentStart == this->start_indices.at(this->start_indices.size()-1))
        segmentEnd = this->run_types.size();
    else
        segmentEnd = this->start_indices.at(si+1);

    segmentLength = segmentEnd - segmentStart;

    firstBreak = this->start_indices.at(si) - si;

    // start and end coordinate of the first run (end coordinate don't belong to the run!)
    startCoordinate = this->min;
    if (segmentLength < 2)  endCoordinate = this->max;
    else endCoordinate = this->run_breaks.at( firstBreak );                             // set endCoordinate for the first run to the first break

    // initialize min and max
    dilatedHRLE->min = this->min;
    dilatedHRLE->max = this->max;

    // dilate the defined runs of this segment and create a new RLE segment
    for (j = segmentStart; j < segmentEnd; j++) {
        run = this->run_types.at(j);                                                    // get run

        if ( run.type == RT_DEFINED ) {
            // dilate the defined run
            dilatedCoordinates.push_back( startCoordinate - pointCount );
            dilatedCoordinates.push_back( endCoordinate + pointCount );
        }

        startCoordinate = endCoordinate;                                                // start coordinate of the next run
        if ((j+2) >= segmentEnd) endCoordinate = this->max;                             // end coordinate of the next run
        else endCoordinate = this->run_breaks.at( ++firstBreak );
    }

    // union of the dilated runs and the original segment
    for (j = 0; j < dilatedCoordinates.size(); j += 2) {

        run_code newRun;

        // start and end coordinate of this defined run
        int definedStartCoordinate = dilatedCoordinates.at(j);
        int definedEndCoordinate = dilatedCoordinates.at(j+1);

        // if the defined run starts before the minimal coordinate
        if ( j == 0 ) {
            if ( definedStartCoordinate <= this->min ) {
                tmpRun = this->getRunByCoordinate( this->min, si );
                if ( tmpRun.type != RT_DEFINED ) {
                    dilatedHRLE->min = definedStartCoordinate - 1;
                    newRun.type = tmpRun.type;
                    dilatedHRLE->run_types.push_back( newRun );
                }
                else dilatedHRLE->min = definedStartCoordinate;
            }
            else dilatedHRLE->run_types.push_back( this->getRunByCoordinate( this->min, si ) );
        }

        // if the run is between min and max
        if ( dilatedHRLE->run_breaks.size() > 0 ) {
            // union of two defined runs
            if (definedStartCoordinate <= dilatedHRLE->run_breaks.back()) {
                dilatedHRLE->run_breaks.pop_back();
            }
            // if a non-defined run lies between the current and the previous defined run
            else {
                newRun = this->getRunByCoordinate( dilatedHRLE->run_breaks.back(), si );
                dilatedHRLE->run_types.push_back( newRun );

                if ( definedStartCoordinate != dilatedHRLE->min ) dilatedHRLE->run_breaks.push_back( definedStartCoordinate );
                newRun.type = RT_DEFINED;
                dilatedHRLE->run_types.push_back( newRun );
            }
            dilatedHRLE->run_breaks.push_back( definedEndCoordinate );
        }
        // add defined run if it is the first defined run in this segment / block
        else {
            if ( definedStartCoordinate != dilatedHRLE->min ) dilatedHRLE->run_breaks.push_back( definedStartCoordinate );
            dilatedHRLE->run_breaks.push_back( definedEndCoordinate );
            newRun.type = RT_DEFINED;
            dilatedHRLE->run_types.push_back( newRun );
        }

        // if the defined run ends behind the maximal coordinate
        if ( (j+2) >= dilatedCoordinates.size() ) {
            if ( definedEndCoordinate >= this->max ) {
                tmpRun = this->getRunByCoordinate( (this->max -1), si );
                if ( tmpRun.type != RT_DEFINED ) {
                    dilatedHRLE->max = definedEndCoordinate + 1;
                    newRun.type = tmpRun.type;
                    dilatedHRLE->run_types.push_back( newRun );
                }
                else dilatedHRLE->max = definedEndCoordinate;
            }
            else dilatedHRLE->run_types.push_back( this->getRunByCoordinate( (this->max -1), si ) );
            if ( (dilatedHRLE->run_breaks.size() > 0) && (dilatedHRLE->run_breaks.back() == dilatedHRLE->max) ) dilatedHRLE->run_breaks.pop_back();
        }

    }

    // start index of this segment is 0 because this dilatedHRLE is an 1D RLE block
    dilatedHRLE->start_indices.push_back(0);

    return dilatedHRLE;
}

/*!
    \fn CHierarchicalRLE::getRunByCoordinate( int qi, int segmentIndex )
 */
// returns the run code of a run in this RLE block by the given coordinate and segment index
run_code CHierarchicalRLE::getRunByCoordinate( int qi, int segmentIndex, int* startCoord ) {

    int                 tmp_min = this->min;
    int                 tmp_max = this->max;
    int                 segmentLength;
    unsigned int        firstRun;
    unsigned int        firstBreak;
    run_code            run;

    // get segment length - number of runs in this segment
    if (this->start_indices.at(segmentIndex) == this->start_indices.at(this->start_indices.size()-1))
        segmentLength = this->run_types.size() - this->start_indices.at(segmentIndex);
    else
        segmentLength = this->start_indices.at(segmentIndex+1) - this->start_indices.at(segmentIndex);

    // get the index of the first run in this segment
    firstRun = this->start_indices.at(segmentIndex);
    // get the index of the first run break of this segment
    firstBreak = this->start_indices.at(segmentIndex) - segmentIndex;

    if ( (qi < this->min) || (qi >= this->max) ) {
        run.type = RT_NEGATIVE;
    }
    else {
        int indexRun = this->accessBinSearch( qi, firstRun, firstBreak, segmentLength, &tmp_min, &tmp_max);
        run = this->run_types.at( indexRun );
    }

    if (startCoord != NULL) *startCoord = tmp_min;

    return run;

}

/*!
    \fn CHierarchicalRLE::addToHRLE( CHierarchicalRLE hrleObject )
 */
// add a hrle structure to the current hrle structure
// hrleObject must be from the same dimension!!
bool CHierarchicalRLE::addToHRLE( CHierarchicalRLE *hrleObject ) {
    int i;
    int numRuns = this->run_types.size();
    run_code run;

    // add the values of the given hrle level set to the current level set - higher dimensions
    if ( (this->lower_hrle != NULL) && (hrleObject->lower_hrle != NULL) ) {
        // add run types and update the start indices
        for ( i = 0; i < hrleObject->run_types.size(); i++ ) {
            run = hrleObject->run_types.at(i);
            if ( run.type == RT_DEFINED )
                run.index += this->lower_hrle->start_indices.size();
            this->run_types.push_back( run );
        }
        // recursive call for the lower hrle block
        if ( this->lower_hrle->addToHRLE( hrleObject->lower_hrle ) == false ) {
            cerr << "Error while adding lower dimension of a HRLE object to another" << endl;
            return false;
        }
    }
    // add the values of the given hrle level set to the current level set - lowest dimension
    else if ( (this->lower_hrle == NULL) && (hrleObject->lower_hrle == NULL) ) {
        // add run types and update the start indices
        for ( i = 0; i < hrleObject->run_types.size(); i++ ) {
            run = hrleObject->run_types.at(i);
            if ( run.type == RT_DEFINED )
                run.index += this->defined_values.size();
            this->run_types.push_back( run );
        }
        // add defined values
        for ( i = 0; i < hrleObject->defined_values.size(); i++ )
            this->defined_values.push_back( hrleObject->defined_values.at(i) );
    }
    else {
        cerr << "Error while adding a hrle level set to another! - Different dimensions!!" << endl;
        return false;
    }

    // add start indices
    for ( i = 0; i < hrleObject->start_indices.size(); i++ )
        this->start_indices.push_back( numRuns + hrleObject->start_indices.at( i ) );
    // add run breaks
    for ( i = 0; i < hrleObject->run_breaks.size(); i++ )
        this->run_breaks.push_back( hrleObject->run_breaks.at(i) );
    // update min & max
    if ( hrleObject->min < this->min ) this->min = hrleObject->min;
    if ( hrleObject->max > this->max ) this->max = hrleObject->max;

    delete hrleObject;
    return true;
}

/*!
    \fn CHierarchicalRLE::getSegmentAt( int si )
 */
// the function is used to retrieve an RLE segment in this dimension with all its subsegments by the segment index
CHierarchicalRLE* CHierarchicalRLE::getSegmentAt( int si ) {
    CHierarchicalRLE* returnValue = new CHierarchicalRLE( this->dimension );
    int firstRun, firstBreak, segmentLength, segmentEnd;
    int startCoord, endCoord;
    int runIndex = 0;
    run_code newRun;
    int definedIndex;
    int i,j;

    // get the index of the first run in this segment
    firstRun = this->start_indices.at( si );
    // get the index of the first run break of this segment
    firstBreak = this->start_indices.at( si ) - si;
    // end of this segment (start of the next one)
    if (this->start_indices.at( si ) == this->start_indices.at(this->start_indices.size()-1))
        segmentEnd = this->run_types.size();
    else
        segmentEnd = this->start_indices.at( si+1 );
    // get segment length - number of runs in this segment
    segmentLength = segmentEnd - firstRun;
    // start- and endcoordinates
    startCoord = this->min;
    if ( segmentLength > 1) endCoord = this->run_breaks.at( firstBreak );
    else endCoord = this->max;

    // set min, max and start index of this segment in the new block
    returnValue->min = this->min;
    returnValue->max = this->max;
    returnValue->start_indices.push_back( 0 );

    if ( this->lower_hrle != NULL ) returnValue->lower_hrle = new CHierarchicalRLE( this->dimension -1 );

    for ( i = firstRun; i < segmentEnd; i++ ) {

        newRun = this->run_types.at( i );

        if (newRun.type == RT_DEFINED ) {

            runIndex = newRun.index;

            if ( this->lower_hrle != NULL )
                newRun.index = returnValue->lower_hrle->start_indices.size();
            else
                newRun.index = returnValue->defined_values.size();

            for ( j = startCoord; j < endCoord; j++ ) {
                definedIndex = runIndex + j - startCoord;
                if ( this->lower_hrle != NULL ) {
                    returnValue->lower_hrle->addToHRLE( this->lower_hrle->getSegmentAt( definedIndex ) );
                }
                else if ( this->defined_values.size() > 0 ) {
                    returnValue->defined_values.push_back( this->defined_values.at( definedIndex ) );
                }
            }

        }

        // add this run
        returnValue->run_types.push_back( newRun );

        // add the break, if it exists
        if ( (i+1) < segmentEnd ) returnValue->run_breaks.push_back( this->run_breaks.at(firstBreak++) );

        startCoord = endCoord;
        if ( (i+2) < segmentEnd ) endCoord = this->run_breaks.at( firstBreak );
        else endCoord = this->max;

    }

    return returnValue;

}

/*!
    \fn CHierarchicalRLE::hrleUnion( vector<CHierarchicalRLE> hrleVector )
 */
// union of some 1D HRLE blocks to one 1D HRLE block
CHierarchicalRLE* CHierarchicalRLE::hrleUnion( vector<CHierarchicalRLE*> hrleVector ) {
    CHierarchicalRLE *tempValue;
    CHierarchicalRLE *returnValue;                               // temporary for the result
    CHierarchicalRLE *currentHRLE;
    int pos1, pos2;
    int startCoord, endCoord;
    int startCoord1, endCoord1, startCoord2, endCoord2;
    run_code runFromOtherBlock;
    int runBreak1, runBreak2;
    run_code newRun;
    vector<CHierarchicalRLE*>       tmpSubsegments;
    run_code lastRun;
    lastRun.type = RT_NEGATIVE;

    // if there is just one HRLE block -> nothing to do, return given HRLE block
    if ( hrleVector.size() == 1 ) {
        return hrleVector.at(0);
    }

    tempValue = hrleVector.at(0);

    // now we need to compare the different HRLE blocks
    for ( int i = 1; i < hrleVector.size(); i++ ) {
        currentHRLE = hrleVector.at(i);                     // the next HRLE block
        pos1 = 0;                                           // index into the run array of the tempValue
        pos2 = 0;                                           // index into the run array of the currentHRLE
        returnValue = new CHierarchicalRLE();
        if ( (tempValue->lower_hrle != NULL) && (currentHRLE->lower_hrle != NULL) ) returnValue->lower_hrle = new CHierarchicalRLE();

        // set dimension
        returnValue->dimension = tempValue->dimension;
        // initialize start indices array
        returnValue->start_indices.push_back( 0 );

        // update min
        if ( currentHRLE->min < tempValue->min ) returnValue->min = currentHRLE->min;
        else returnValue->min = tempValue->min;

        // update max
        if ( currentHRLE->max > tempValue->max ) returnValue->max = currentHRLE->max;
        else returnValue->max = tempValue->max;

        // initialize start and end coordinates (first run)
        startCoord1 = tempValue->min;
        startCoord2 = currentHRLE->min;
        if ( tempValue->run_types.size() < 2 ) endCoord1 = tempValue->max;
        else endCoord1 = tempValue->run_breaks.at( 0 );
        if ( currentHRLE->run_types.size() < 2 ) endCoord2 = currentHRLE->max;
        else endCoord2 = currentHRLE->run_breaks.at( 0 );

        // traverse through all runs of both H-RLE blocks
        while ( ( pos1 < tempValue->run_types.size()) || (pos2 < currentHRLE->run_types.size()) ) {

            // get the breaks of both runs
            if ( pos1 >= tempValue->run_breaks.size() ) runBreak1 = tempValue->max;
            else runBreak1 = tempValue->run_breaks.at(pos1);
            if ( pos2 >= currentHRLE->run_breaks.size() ) runBreak2 = currentHRLE->max;
            else runBreak2 = currentHRLE->run_breaks.at(pos2);

            // add the remaining runs and breaks of currentHRLE
            if ( (pos1 >= tempValue->run_types.size()) && (pos2 < currentHRLE->run_types.size()) && (runBreak2 > returnValue->run_breaks.back()) ) {
                for ( int i = pos2; i < currentHRLE->run_types.size(); i++ ) {
                    newRun.type = currentHRLE->run_types[i].type;

                    if ( returnValue->run_breaks.size() > 0 ) startCoord = returnValue->run_breaks.back();
                    else startCoord = returnValue->min;

                    if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                            returnValue->run_breaks.pop_back();
                    }

                    // if the run is defined, add the subsegments to the result
                    if ( newRun.type == RT_DEFINED ) {
                        if ( currentHRLE->lower_hrle != NULL ) newRun.index = returnValue->lower_hrle->start_indices.size();

                        if ( i < currentHRLE->run_breaks.size() ) endCoord = currentHRLE->run_breaks.at(i);
                        else endCoord = currentHRLE->max;

                        if ( currentHRLE->lower_hrle != NULL ) {
                            for ( int qi = startCoord; qi < endCoord; qi++) {
                                returnValue->lower_hrle->addToHRLE( currentHRLE->lower_hrle->getSegmentAt( currentHRLE->run_types[pos2].index + qi - startCoord ) );
                            }
                        }
                    }

                    if ( (i < currentHRLE->run_breaks.size()) && (currentHRLE->run_breaks.at(i) != returnValue->max) &&
                        ( (returnValue->run_breaks.size() == 0) || (currentHRLE->run_breaks.at(i) > returnValue->run_breaks.back()) ) &&
                        (newRun.type != lastRun.type) )
                        returnValue->run_breaks.push_back( currentHRLE->run_breaks.at(i) );

                    if ( (returnValue->run_types.size() == 0) || (lastRun.type != newRun.type) )
                        returnValue->run_types.push_back( newRun );

                    lastRun.type = newRun.type;
                }
                pos2 = currentHRLE->run_types.size();
            }
            // add the remaining runs and breaks of tempValue
            else if ( (pos2 >= currentHRLE->run_types.size()) && (pos1 < tempValue->run_types.size()) && (runBreak1 > returnValue->run_breaks.back()) ) {
                for ( int i = pos1; i < tempValue->run_types.size(); i++ ) {
                    newRun.type = tempValue->run_types[i].type;

                    if ( returnValue->run_breaks.size() > 0 ) startCoord = returnValue->run_breaks.back();
                    else startCoord = returnValue->min;

                    if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                            returnValue->run_breaks.pop_back();
                    }

                    // if the run is defined, add the subsegments to the result
                    if ( newRun.type == RT_DEFINED ) {
                        if ( tempValue->lower_hrle != NULL ) newRun.index = returnValue->lower_hrle->start_indices.size();

                        if ( i < tempValue->run_breaks.size() ) endCoord = tempValue->run_breaks.at(i);
                        else endCoord = currentHRLE->max;

                        if ( tempValue->lower_hrle != NULL ) {
                            for ( int qi = startCoord; qi < endCoord; qi++) {
                                returnValue->lower_hrle->addToHRLE( tempValue->lower_hrle->getSegmentAt( tempValue->run_types[pos1].index + qi - startCoord ) );
                            }

                        }                    }

                    if ( (i < tempValue->run_breaks.size()) && (tempValue->run_breaks.at(i) != returnValue->max) &&
                        ( (returnValue->run_breaks.size() == 0) || (tempValue->run_breaks.at(i) > returnValue->run_breaks.back()) ) &&
                        (newRun.type != lastRun.type) )
                        returnValue->run_breaks.push_back( tempValue->run_breaks.at(i) );

                    if ( (returnValue->run_types.size() == 0) || (lastRun.type != newRun.type) )
                        returnValue->run_types.push_back( newRun );

                    lastRun.type = newRun.type;
                }
                pos1 = tempValue->run_types.size();
            }
            // compare the runs of both HRLE blocks
            else if ( ( pos1 < tempValue->run_types.size()) && (pos2 < currentHRLE->run_types.size())) {

                if ( (returnValue->run_breaks.size() == 0) || ((runBreak1 > returnValue->run_breaks.back()) && (runBreak2 > returnValue->run_breaks.back())) ) {
                    // if both runs have the same value
                    if ( tempValue->run_types[pos1].type == currentHRLE->run_types[pos2].type ) {

                        newRun.type = tempValue->run_types[pos1].type;

                        if ( returnValue->run_breaks.size() > 0 ) startCoord = returnValue->run_breaks.back();
                        else startCoord = returnValue->min;

                        // if the last run has the same type, don't add a new run
                        if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                            returnValue->run_breaks.pop_back();
                        }

                        // if defined, use the higher break and get the subsegments
                        if ( newRun.type == RT_DEFINED ) {

                            if ( runBreak1 > runBreak2 ) returnValue->run_breaks.push_back( runBreak1 );
                            else returnValue->run_breaks.push_back( runBreak2 );

                            endCoord = returnValue->run_breaks.back();

                            if ( tempValue->lower_hrle != NULL ) {

                                newRun.index = returnValue->lower_hrle->start_indices.size();

                                for ( int qi = startCoord; qi < endCoord; qi++) {

                                    tmpSubsegments.clear();

                                    // get subsegments of the run from the first H-RLE block
                                    if ( qi < endCoord1 )
                                        tmpSubsegments.push_back( tempValue->lower_hrle->getSegmentAt( tempValue->run_types[pos1].index + qi - startCoord1 ) );         // subsegment of the first hrle
                                    else if ( qi < tempValue->max ) {
                                        int tmpStartOtherRun = tempValue->min;
                                        runFromOtherBlock = tempValue->getRunByCoordinate( qi, 0, &tmpStartOtherRun );
                                        if ( runFromOtherBlock.type == RT_DEFINED )
                                            tmpSubsegments.push_back( tempValue->lower_hrle->getSegmentAt( runFromOtherBlock.index + qi - tmpStartOtherRun ) );
                                    }

                                    // get subsements of the run from the second H-RLE block
                                    if ( qi < endCoord2 )
                                        tmpSubsegments.push_back( currentHRLE->lower_hrle->getSegmentAt( currentHRLE->run_types[pos2].index + qi - startCoord2 ) );     // subsegment of the second hrle
                                    else  if ( qi < currentHRLE->max ) {
                                        int tmpStartOtherRun = currentHRLE->min;
                                        runFromOtherBlock = currentHRLE->getRunByCoordinate( qi, 0, &tmpStartOtherRun );
                                        if ( runFromOtherBlock.type == RT_DEFINED )
                                            tmpSubsegments.push_back( currentHRLE->lower_hrle->getSegmentAt( runFromOtherBlock.index + qi - tmpStartOtherRun ) );
                                    }

                                    if ( tmpSubsegments.size() >= 1 )
                                        returnValue->lower_hrle->addToHRLE( returnValue->hrleUnion( tmpSubsegments ) );     // add the union of these two segments to the resulting hrle
                                    else if ( tmpSubsegments.size() < 1 )
                                        cout << "Error in function union -> at both defined" << endl;
                                }
                            }
                        }
                        // if undefined, use the lower break
                        else {
                            if ( runBreak1 < runBreak2 ) returnValue->run_breaks.push_back( runBreak1 );
                            else returnValue->run_breaks.push_back( runBreak2 );
                        }

                        if ( (returnValue->run_types.size() == 0) || (lastRun.type != newRun.type) ) {
                            // add a new run
                            returnValue->run_types.push_back( newRun );
                        }

                    }
                    // else choose at first defined run, than inner run and at last outer run
                    else {

                        if ( returnValue->run_breaks.size() > 0 ) startCoord = returnValue->run_breaks.back();
                        else startCoord = returnValue->min;

                        if ( tempValue->run_types[pos1].type == RT_DEFINED ) {

                            newRun.type = RT_DEFINED;

                            // same run so we don't need the last break
                            if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                                returnValue->run_breaks.pop_back();
                            }
                            returnValue->run_breaks.push_back( runBreak1 );
                            endCoord = returnValue->run_breaks.back();

                            if ( tempValue->lower_hrle != NULL ) {

                                newRun.index = returnValue->lower_hrle->start_indices.size();

                                for ( int qi = startCoord; qi < endCoord; qi++) {

                                    tmpSubsegments.clear();

                                    int tmpStartOtherRun = currentHRLE->min;

                                        tmpSubsegments.push_back( tempValue->lower_hrle->getSegmentAt( tempValue->run_types[pos1].index + qi - startCoord1 ) );

                                    if ( qi < currentHRLE->max ) {
                                        runFromOtherBlock = currentHRLE->getRunByCoordinate( qi, 0, &tmpStartOtherRun );
                                        if ( runFromOtherBlock.type == RT_DEFINED )
                                            tmpSubsegments.push_back( currentHRLE->lower_hrle->getSegmentAt( runFromOtherBlock.index + qi - tmpStartOtherRun ) );
                                    }

                                    if ( tmpSubsegments.size() >= 1 )
                                        returnValue->lower_hrle->addToHRLE( returnValue->hrleUnion( tmpSubsegments ) );     // add the union of these two segments to the resulting hrle
                                    else if ( tmpSubsegments.size() < 1 )
                                        cout << "Error in function union -> at different types; first defined" << endl;

                                }
                            }
                        }
                        else if ( currentHRLE->run_types[pos2].type == RT_DEFINED ) {

                            newRun.type = RT_DEFINED;

                            // same run so we don't need the last break
                            if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                                returnValue->run_breaks.pop_back();
                            }

                            returnValue->run_breaks.push_back( runBreak2 );
                            endCoord = returnValue->run_breaks.back();

                            if ( currentHRLE->lower_hrle != NULL ) {

                                newRun.index = returnValue->lower_hrle->start_indices.size();

                                for ( int qi = startCoord; qi < endCoord; qi++) {

                                    tmpSubsegments.clear();

                                    int tmpStartOtherRun = tempValue->min;

                                    tmpSubsegments.push_back( currentHRLE->lower_hrle->getSegmentAt( currentHRLE->run_types[pos2].index + qi - startCoord2 ) );

                                    if ( qi < tempValue->max ) {
                                        runFromOtherBlock = tempValue->getRunByCoordinate( qi, 0, &tmpStartOtherRun );
                                        if ( runFromOtherBlock.type == RT_DEFINED )
                                            tmpSubsegments.push_back( tempValue->lower_hrle->getSegmentAt( runFromOtherBlock.index + qi - tmpStartOtherRun ) );
                                    }

                                    if ( tmpSubsegments.size() >= 1 )
                                        returnValue->lower_hrle->addToHRLE( returnValue->hrleUnion( tmpSubsegments ) );     // add the union of these two segments to the resulting hrle
                                    else if ( tmpSubsegments.size() < 1 )
                                        cout << "Error in function union -> at different types; first defined" << endl;
                                }
                            }
                        }
                        // first run is an inner run, second an outer
                        else if ( tempValue->run_types[pos1].type == RT_NEGATIVE ) {
                            newRun.type = RT_NEGATIVE;
                            if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                                returnValue->run_breaks.pop_back();
                            }
                            if ( runBreak1 < runBreak2 ) returnValue->run_breaks.push_back( runBreak1 );
                            else returnValue->run_breaks.push_back( runBreak2 );
                        }
                        // second run is an inner run, first an outer
                        else if ( currentHRLE->run_types[pos2].type == RT_NEGATIVE ) {
                            newRun.type = RT_NEGATIVE;
                            if ( (returnValue->run_types.size() > 0) && (lastRun.type == newRun.type) ) {
                                returnValue->run_breaks.pop_back();
                            }
                            if ( runBreak1 < runBreak2 ) returnValue->run_breaks.push_back( runBreak1 );
                            else returnValue->run_breaks.push_back( runBreak2 );
                        }
                        // well ... that shouldn't happen :)
                        else {
                            cout << "Error during union!" << endl;
                            cout << "tempValue Type: " << tempValue->run_types[pos1].type << endl;
                            cout << "currentHRLE Type: " << currentHRLE->run_types[pos2].type << endl;
                        }

                        if ( (returnValue->run_types.size() == 0) || (lastRun.type != newRun.type) ) {
                            // add a new run
                            returnValue->run_types.push_back( newRun );
                        }
                    }
                }
            }
            else {
                pos1 = tempValue->run_types.size();
                pos2 = currentHRLE->run_types.size();
            }

            // update pos -> next run if the last break in the result is greater then the break of this run
            // if the break of the run in tempValue is greater as the last break in the result, go to the next run in the currentHRLE
            if ( ((runBreak1 <= returnValue->run_breaks.back()) && (runBreak2 <= returnValue->run_breaks.back()))) {
                pos1++;
                pos2++;

                // update start and end coordinate of the next run
                startCoord1 = endCoord1;
                startCoord2 = endCoord2;
                if ( pos1 < tempValue->run_breaks.size()) endCoord1 = tempValue->run_breaks.at(pos1);
                else endCoord1 = tempValue->max;
                if ( pos2 < currentHRLE->run_breaks.size()) endCoord2 = currentHRLE->run_breaks.at(pos2);
                else endCoord2 = currentHRLE->max;
            }
            else if ( (runBreak1 > returnValue->run_breaks.back()) && (runBreak2 <= returnValue->run_breaks.back()) ) {
                pos2++;

                // update start and end coordinate of the next run
                startCoord2 = endCoord2;
                if ( pos2 < currentHRLE->run_breaks.size()) endCoord2 = currentHRLE->run_breaks.at(pos2);
                else endCoord2 = currentHRLE->max;
            }
            else if ( (runBreak2 > returnValue->run_breaks.back()) && (runBreak1 <= returnValue->run_breaks.back()) ) {
                pos1++;

                // update start and end coordinate of the next run
                startCoord1 = endCoord1;
                if ( pos1 < tempValue->run_breaks.size()) endCoord1 = tempValue->run_breaks.at(pos1);
                else endCoord1 = tempValue->max;
            }

            lastRun = returnValue->run_types.back();

        }

        delete currentHRLE;
        delete tempValue;
        tempValue = returnValue;

        if ( (tempValue->run_breaks.size() > 0) && (tempValue->run_breaks.back() == tempValue->max) ) tempValue->run_breaks.pop_back();

    }

    return tempValue;
}

/*!
    \fn CHierarchicalRLE::updateValueArray( CHierarchicalRLE* hrleObject )
 */
// update the value array after dilation
bool CHierarchicalRLE::updateValueArray( CHierarchicalRLE* originalHrle, vector<int> *parentCoord, int si ) {

    int runIndex, definedIndex;
    int qi, startCoord, endCoord;
    int segmentStart, segmentEnd, segmentLength;
    run_code tmpRun, originalRun;
    vector<int> tmpVector;

    if ( parentCoord == NULL ) parentCoord = new vector<int>;

    // start and end index of this segment
    segmentStart = this->start_indices.at( si );
    if ( (si+1) < this->start_indices.size() ) segmentEnd = this->start_indices.at( si+1 );
    else segmentEnd = this->run_types.size();

    // segment length
    segmentLength = segmentEnd - segmentStart;

    // start and end coordinate of the first run
    startCoord = this->min;
    if ( segmentLength > 1 ) endCoord = this->run_breaks.at(segmentStart - si);
    else endCoord = this->max;

    // for all runs in this segment
    for ( runIndex = segmentStart; runIndex < segmentEnd; runIndex++ ) {
        tmpRun = this->run_types.at( runIndex );

        if ( tmpRun.type == RT_DEFINED ) {

            if ( this->lower_hrle == NULL ) this->run_types[runIndex].index = this->defined_values.size();

            // if the run is defined, iterate over its coordinates
            for ( qi = startCoord; qi < endCoord; qi++ ) {

                // update the parent coordinates
                parentCoord->push_back( qi );

                if ( this->lower_hrle != NULL ) {
                    definedIndex = tmpRun.index + qi - startCoord;
                    this->lower_hrle->updateValueArray( originalHrle, parentCoord, definedIndex );
                }
                else {

                        tmpVector.clear();

                        for ( int i = (parentCoord->size()-1); i >= 0; i-- )
                            tmpVector.push_back( parentCoord->at(i) );

                        originalRun = originalHrle->randomAccess( tmpVector );

                        if ( originalRun.type == RT_DEFINED )
                            this->defined_values.push_back( originalHrle->getDefinedValue( originalRun.index ) );
                        else
                            this->defined_values.push_back( 1 );

                }

                parentCoord->pop_back();
            }
        }

        startCoord = endCoord;
        if ( (runIndex+2) < segmentEnd ) endCoord = this->run_breaks.at( runIndex +1 -si );
        else endCoord = this->max;
    }


    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//      Methodes for testing only
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CHierarchicalRLE::draw()
 */
// draw the level set as cubes - only for 3d level sets implemented
void CHierarchicalRLE::drawCubes()
{
    // for comnputing the average access time of the random access operator
    long            numGridPoints = 0, accessTime = 0, totalAccessTime = 0, averageAccessTime = 0;
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)

    if (this->dimension == 3) {

        int minX = this->lower_hrle->lower_hrle->min;
        int maxX = this->lower_hrle->lower_hrle->max;
        int minY = this->lower_hrle->min;
        int maxY = this->lower_hrle->max;
        int minZ = this->min;
        int maxZ = this->max;

        float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
        float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
        float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

        for (int z = minZ; z < maxZ; z++) {
            for (int y = minY; y < maxY; y++) {
                for (int x = minX; x < maxX; x++) {
                    numGridPoints++;

                    gettimeofday( &tp1, NULL );
                    run_code run = this->randomAccess( x, y, z );
                    gettimeofday( &tp2, NULL );

                    accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);
                    totalAccessTime += accessTime;

                    if (run.type == RT_DEFINED) {

                        // draw cubes and cube edges
                        // -------------------------

                        CVertex3D point1( this->minRange.getX() + x * CubeW,
                                          this->minRange.getY() + y * CubeH,
                                          this->minRange.getZ() + z * CubeD );
                        CVertex3D point2( this->minRange.getX() + (x+1) * CubeW,
                                          this->minRange.getY() + (y+1) * CubeH,
                                          this->minRange.getZ() + (z+1) * CubeD);

                        this->drawACube(&point1, &point2, this->lower_hrle->lower_hrle->defined_values.at(run.index));

                    }
                }
            }
        }

    }

    averageAccessTime = totalAccessTime / numGridPoints;

    cout << "Average access time for the Random Access Operator  while drawing the whole object was " << averageAccessTime << " micro seconds." << endl;

    this->accessTimeRandomAccess = (double) accessTime;
    this->accessAverageTimeRandomAccess = (double) totalAccessTime / (double) numGridPoints;
}

/*!
    \fn CHierarchicalRLE::draw()
 */
// draw the level set as cubes (using the sequential access) - only for 3d level sets implemented
void CHierarchicalRLE::drawCubesUsingSequentialAccess()
{
    long            accessTime = 0;     // the access time for the function
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)

    if (this->dimension == 3) {

        int minX = this->lower_hrle->lower_hrle->min;
        int maxX = this->lower_hrle->lower_hrle->max;
        int minY = this->lower_hrle->min;
        int maxY = this->lower_hrle->max;
        int minZ = this->min;
        int maxZ = this->max;

        float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
        float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
        float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

        // get all defined voxels
        gettimeofday( &tp1, NULL );
        vector<coordinate_value> definedCoordinates = this->sequentialAccess();
        gettimeofday( &tp2, NULL );

        for (int i = 0; i < definedCoordinates.size(); i++) {

            vector<int> coordinate = definedCoordinates[i].coordinates;
            int index = definedCoordinates[i].index;

            CVertex3D point1( this->minRange.getX() + coordinate[2] * CubeW,
                              this->minRange.getY() + coordinate[1] * CubeH,
                              this->minRange.getZ() + coordinate[0] * CubeD );
            CVertex3D point2( this->minRange.getX() + (coordinate[2]+1) * CubeW,
                              this->minRange.getY() + (coordinate[1]+1) * CubeH,
                              this->minRange.getZ() + (coordinate[0]+1) * CubeD);

            this->drawACube(&point1, &point2, this->lower_hrle->lower_hrle->defined_values.at(index));

        }

        accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);

    }

    cout << "Access time for the Sequential Access Operator was " << accessTime << " micro seconds." << endl;
    this->accessTimeSequentialAccess = (double) accessTime;
}

void CHierarchicalRLE::drawACube(CVertex3D *p1, CVertex3D *p2, double colorValue) {

    // Fl�hen zeichnen
    // ----------------
        // Farbe festlegen
        glColor3f(0.0, 0.0, 1.0 * colorValue);

        // Offset fr den Cube (bessere Darstellung der Kanten)
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(1.f, 1.f);

            glBegin(GL_POLYGON);
                glNormal3f(  -1.0,  0.0,  0.0);
                glVertex3f(  p1->getX(), p1->getY(), p1->getZ());
                glVertex3f(  p1->getX(), p1->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d( 1.0,  0.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  -1.0,  0.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0, 1.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, -1.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, 1.0);
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

        glDisable(GL_POLYGON_OFFSET_FILL);


// Kanten zeichnen
// ---------------

        glColor3d(0.2, 0.2, 0.2);

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();


    return;
}

CHierarchicalRLE* CHierarchicalRLE::getDilatedHRLE( int gridPoints ) {

    CHierarchicalRLE* dilatedHRLE = this->dilation( gridPoints );
    dilatedHRLE->minRange = this->minRange;
    dilatedHRLE->maxRange = this->maxRange;
    dilatedHRLE->lower_hrle->minRange = this->lower_hrle->minRange;
    dilatedHRLE->lower_hrle->maxRange = this->lower_hrle->maxRange;
    dilatedHRLE->lower_hrle->lower_hrle->minRange = this->lower_hrle->lower_hrle->minRange;
    dilatedHRLE->lower_hrle->lower_hrle->maxRange = this->lower_hrle->lower_hrle->maxRange;
    return dilatedHRLE;

}

CHierarchicalRLE* CHierarchicalRLE::removeFarPoints( double distance ) {

    CHierarchicalRLE* tmp = this->removeFarGridPoints( distance );
    if (tmp == NULL) cerr << "Error while removing far points." << endl;
//     CHierarchicalRLE* tmp = new CHierarchicalRLE(3);
    tmp->minRange = this->minRange;
    tmp->maxRange = this->maxRange;
    tmp->lower_hrle->minRange = this->lower_hrle->minRange;
    tmp->lower_hrle->maxRange = this->lower_hrle->maxRange;
    tmp->lower_hrle->lower_hrle->minRange = this->lower_hrle->lower_hrle->minRange;
    tmp->lower_hrle->lower_hrle->maxRange = this->lower_hrle->lower_hrle->maxRange;
    return tmp;

}


void CHierarchicalRLE::printDataStructure() {

    cout << "Ausgabe der Datenstruktur, Dimension: " << this->dimension << endl;
    int i;
    cout << "Min: " << this->min << endl;
    cout << "Max: " << this->max << endl;
    cout << "Start Indices (" << this->start_indices.size() << "): ";
    for (i = 0; i < this->start_indices.size(); i++) cout << this->start_indices.at(i) << " | ";
    cout << endl;
    cout << "Run Types (" << this->run_types.size() << "): ";
    for (i = 0; i < this->run_types.size(); i++) {
        if ( this->run_types[i].type == RT_DEFINED ) cout << "d "<< this->run_types[i].index <<" | ";
        if ( this->run_types[i].type == RT_NEGATIVE ) cout << "- | ";
        if ( this->run_types[i].type == RT_POSITIVE ) cout << "+ | ";
    }
    cout << endl;
    cout << "Run Breaks (" << this->run_breaks.size() << "): ";
    for (i = 0; i < this->run_breaks.size(); i++) cout << this->run_breaks.at(i) << " | ";
    cout << endl;

    if (this->lower_hrle != NULL) this->lower_hrle->printDataStructure();
    else {
        cout << "Defined Values (" << this->defined_values.size() << "): ";
        for (i = 0; i < this->defined_values.size(); i++) cout << this->defined_values.at(i) << " | ";
        cout << endl;
    }

}


/*!
    \fn CHierarchicalRLE::printNeighborOfGridPoint(int x, int y, int z, int axis = 0)
 */
void CHierarchicalRLE::printNeighborOfGridPoint(int x, int y, int z, int axis)
{

    cout << "Neighbor search of grid point (" << x << "," << y << "," << z << ") on";
    if ( axis == 3 ) cout << " encoding axis Z." << endl;
    if ( axis == 2 ) cout << " encoding axis Y." << endl;
    if ( axis == 1 ) cout << " encoding axis X." << endl;
    if ( axis == 0 ) cout << " all encoding axis.." << endl;

    vector<neighbors> returnedNeighbors = this->neighborAccess( x, y, z, axis );

    cout << "Returned Neighbors Size: " << returnedNeighbors.size() << endl;

    if ( returnedNeighbors.size() > 0 ) {
        for ( int i = 0; i < returnedNeighbors.size(); i++ ) {
            neighbors nb = returnedNeighbors.at(i);

            cout << "The grid point has a neighbor on the ";
            if ( nb.dimension == 3 ) cout << " encoding axis Z";
            if ( nb.dimension == 2 ) cout << " encoding axis Y";
            if ( nb.dimension == 1 ) cout << " encoding axis X";
            if ( nb.leftNeighbor ) cout << " in negative coordinate direction with value " << nb.leftValue;
            if ( nb.leftNeighbor && nb.rightNeighbor ) cout << " and";
            if ( nb.rightNeighbor ) cout << " in positive coordinate direction with value " << nb.rightValue;
            cout << "." << endl;
        }
    }
    else {
        cout << "Grid point (" << x << "," << y << "," << z << ") has no neighbors";
        if (axis == 0) cout << "." << endl;
        else cout << " on the given axis." << endl;
    }

}


/*!
    \fn CHierarchicalRLE::getMemorySize()
 */
// returns the size of the data structure in bytes
double CHierarchicalRLE::getMemorySize()
{
    double returnValue;

    returnValue += this->start_indices.size() * sizeof(unsigned int);
    returnValue += 2* sizeof(int);                                      // min and max
    returnValue += this->run_breaks.size() * sizeof(int);
    returnValue += this->defined_values.size() * sizeof(double);
    returnValue += this->run_types.size() * sizeof(run_code);

    // add the used memory of the lower dimensions
    if ( this->lower_hrle != NULL) returnValue += this->lower_hrle->getMemorySize();

    return returnValue;
}
