/***************************************************************************
 *   Implementation of the Hierarchical RLE level set                      *
 *   Copyright (C) 2006 by Holger Gerth                                    *
 *   Chemnitz University of Technology                                     *
 *   holger.gerth@informatik.tu-chemnitz.de                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef HIERARCHICALRLE_H
#define HIERARCHICALRLE_H

#include <vector>
#include <stdexcept>
#include "includes.h"
#include "voxelobject3d.h"
#include "structures.h"

/**
    * @short Hierarchical Run Lenght Encoded Level Set
    * @author Holger Gerth <holger.gerth@informatik.tu-chemnitz.de>
    * @version 0.1
*/

class CHierarchicalRLE{
public:
    vector<unsigned int>            start_indices;          // start indices and extends for the segments
    int                             min;                    // minimal extend along the axis of encoding
    int                             max;                    // maximum+1 extend along the axis of encoding
    vector<int>                     run_breaks;             // start of each run (exept of the first run)
    vector<double>                  defined_values;         // values of defined voxels
    vector<run_code>                run_types;              // how each segment is divided into different runs
    CHierarchicalRLE*               lower_hrle;             // hrle level set of dimension-1

    CHierarchicalRLE();
    CHierarchicalRLE( unsigned int dim );
    CHierarchicalRLE( unsigned int dim, CVoxelObject3D* vo );

    ~CHierarchicalRLE();

    // public classes
    bool convert_to_hrle( CVoxelObject3D* vo );                                                 // convert an 3D Voxel Grid into a HRLE level set
    double getDefinedValue( int dataIndex );                                                    // returns the defined value by the given index
    run_code randomAccess( vector<int> qv, int segmentIndex = 0);                               // random access operator
    run_code randomAccess( int X, int Y, int Z, int segmentIndex = 0);
    vector<neighbors> neighborAccess( vector<int> qv, int axis = 0, int segmentIndex = 0);      // operator for neighbor access
    vector<neighbors> neighborAccess( int X, int Y, int Z, int axis = 0, int segmentIndex = 0);
    vector<coordinate_value> sequentialAccess();                                                // sequential access operator
    vector<coordinate_value> sequentialAccess( vector<int> parentCoord, int segmentIndex = 0 );
    CHierarchicalRLE* rebuild ( double distance );                                              // returns the rebuilded HRLE level set

///////// for testing only /////////////
        // for time measurement
        double accessTimeRandomAccess, accessAverageTimeRandomAccess, accessTimeSequentialAccess;
        // for draw cubes
        CVertex3D                       minRange, maxRange;                  // range

        void drawCubesUsingSequentialAccess();                               // for testing sequential access
        void drawCubes();                                                    // for testing random access
        CHierarchicalRLE* getDilatedHRLE( int gridPoints );                  // returns the dilated HRLE level set
        CHierarchicalRLE* removeFarPoints( double distance );                // returns the HRLE level set without points which are further away than distance
        void printDataStructure();                                           // prints the full data structure to std out
        void printNeighborOfGridPoint(int x, int y, int z, int axis = 0);    // print the values of the neighbors of grid point (x,y,z)
        double getMemorySize();
///////////////////////////////////////

private:
    unsigned int                    dimension;                           // dimension of the hrle level set

    bool create_hrle_from_voxel_object( CVoxelObject3D* vo );            // transform voxel object into an hrle level set
//     void encodeRun( CVoxelObject3D *vo, CUBE_TYPE type );                // encode a single run
    int accessBinSearch( int qi, unsigned int indexFirstRun, unsigned int indexFirstBreak, int segmentLength, int *min, int *max );       // binary search for the random access operator
    CHierarchicalRLE* dilation( int pointCount );                        // dilation algorithm for rebuilding
    CHierarchicalRLE* dilation( CHierarchicalRLE* dilatedHRLE, int pointCount, int si = 0 );
    CHierarchicalRLE* dilateSegment( int si, int pointCount );           // dilate one 1D HRLE segment
    run_code getRunByCoordinate( int qi, int segmentIndex, int* startCoord = NULL );    // get run type by coordinate & segment index
    bool addToHRLE( CHierarchicalRLE *hrleObject );                      // add another HRLE Level Set to the current
    CHierarchicalRLE* hrleUnion( vector<CHierarchicalRLE*> hrleVector ); // union of some HRLE Level Sets
    bool updateValueArray( CHierarchicalRLE* originalHrle, vector<int> *parentCoord = NULL, int si = 0 );
    CHierarchicalRLE* getSegmentAt( int si );                            // return this segment as seperate H-RLE 
    CHierarchicalRLE* removeFarGridPoints( double distance, int segmentIndex = 0 );     // remove all grid points which are further away from the interface than distance

///////// for testing only /////////////
        void drawACube(CVertex3D *p1, CVertex3D *p2, double colorValue);    // draw the voxels as cubes
///////////////////////////////////////
};

#endif
