/***************************************************************************
 *   Lightwave Object Parser                                               *
 *   Version 0.1                                                           *
 *   Copyright (C) 2005 by Holger Gerth                                    *
 *   holger.gerth@informatik.tu-chemnitz.de                                *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "lightwaveobject.h"

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CLightwaveObject::CLightwaveObject()
{

    this->iFaceCount = 0;
    this->iVertexCount = 0;
    this->vBB_upperCorner = CVertex3D();
    this->vBB_lowerCorner = CVertex3D();

}

CLightwaveObject::~CLightwaveObject()
{

}

///////////////////////////////////////////////////////////////////////////////////////////////
// Load Lightwave Objects
///////////////////////////////////////////////////////////////////////////////////////////////

bool CLightwaveObject::loadObjectFile(const char* filename) {

    char cStr[5];                                // temporary variable for the Chunk-Type (4 byte character)

    cStr[4] = '\0';

    unsigned long iFileChunkSize = 0;            // temporary variable for the Chunk Size of the whole file (4 byte integer)
    unsigned long iChunkSize = 0;                // Chunk Size of one chunk (4 byte integer)

    // open .lwo file
    ifstream file(filename,ios::binary);
    if (!file) { 
        cerr << "Error. Cannot open file " << filename << " !!!";
        return false;
    }

    // read the file content
    if (!file.good()) {
        cerr << "Error while opening file " << filename << " !!!";
        return false;
    }
    else {

        // pass the FORM tag
        file.read(cStr,4);
        if (strcmp(cStr, "FORM")) {
            cerr << "Error. False file format! - No IFF File Format." << endl;
            return false;
        }
        else cout << "Parsing IFF File Format." << endl;

        // read the file chunk size (file size minus 8 bytes)
        file.read((char*)&iFileChunkSize,sizeof(iFileChunkSize));
        iFileChunkSize = littleEndian2BigEndian(iFileChunkSize);
        cout << "Chunksize of this File: " << iFileChunkSize << endl;

        // pass the the next 4 byte; this is the filetype (LWO2) 
        file.read(cStr,4);
        iFileChunkSize -= 4;
        if (strcmp(cStr, "LWO2")) {
            cerr << "Error. False file format! - File Type must be LWO2!" << endl;
            return false;
        }
        else cout << "File type: " << cStr << endl;

    }

/****************  Begin reading file content  ****************/
    while (file.good() && (iFileChunkSize > 0)) {

        // read the next tag
        file.read(cStr,4);

        // read the chunk size of this tag
        file.read((char*)&iChunkSize,sizeof(iChunkSize));
        iChunkSize = littleEndian2BigEndian(iChunkSize);

        cout << "Parsing: " << cStr << " - Chunksize: " << iChunkSize << endl;;

        // TAGS - this is a list of surface names
        if (!strcmp(cStr, "TAGS")) {
            this->loadSurfaceNames(&file, iChunkSize);
        }

        // LAYR - if multiple layers are used
        else if (!strcmp(cStr, "LAYR")) {
            /// @TODO implement function for multiple layers
            file.ignore(iChunkSize);			// currently not implemented
        }

        // PNTS - vertex data
        else if (!strcmp(cStr, "PNTS")) {
            this->loadVertices(&file, iChunkSize);
        }

        // BBOX - Bounding Box of the whole object
        else if (!strcmp(cStr, "BBOX")) {
            this->loadBoundingBox(&file, iChunkSize);
        }

        // POLS - polygons of the object
        else if (!strcmp(cStr, "POLS")) {
            this->loadPolygons(&file, iChunkSize);
        }

        // PTAG - index table for connection between surfaces (color) and polygons
        else if (!strcmp("PTAG",cStr)) {

            // type of the PTAG
            file.read(cStr,4);

            cout << "         Type of PTAG: " << cStr << endl;

            // load the contents of this PTAG chunk
            this->loadPTAG(cStr, iChunkSize-4, &file);

        }

        // SURF - Surface Chunks
        else if (!strcmp("SURF",cStr)) {
            // load this surface
            this->loadSurface(iChunkSize, &file);
        }

        // ignore all other data
        else {
            file.ignore(iChunkSize);
        }


        iFileChunkSize -= (iChunkSize + 8); 

    }

    // close the file
    file.close();

    return true;

}

/////////////////////////////////////////////////////////////////////
// convert LittleEndian Byteorder to BigEndian Byteorder
/////////////////////////////////////////////////////////////////////

unsigned long CLightwaveObject::littleEndian2BigEndian(unsigned long original) {

    unsigned long converted;

    converted = 0;

    if (original == 0) return converted;

    converted |= (original & 0xff000000) >> 24;
    converted |= (original & 0x00ff0000) >> 8;
    converted |= (original & 0x0000ff00) << 8;
    converted |= (original & 0x000000ff) << 24;

    return converted;

}

unsigned short int CLightwaveObject::littleEndian2BigEndian(unsigned short int original) {

    unsigned short int converted;

    converted = 0;

    if (original == 0) return converted;

    converted |= (original & 0xff00) >> 8;
    converted |= (original & 0x00ff) << 8;

    return converted;

}

char* CLightwaveObject::littleEndian2BigEndian(char* original) {

    char* converted = NULL;

    if (original == NULL) return converted;

    *converted |= (*original & 0xff000000) >> 24;
    *converted |= (*original & 0x00ff0000) >> 8;
    *converted |= (*original & 0x0000ff00) << 8;
    *converted |= (*original & 0x000000ff) << 24;


    return converted;

}

/////////////////////////////////////////////////////////////////////
// private functions to load some contents from the lwo-file
/////////////////////////////////////////////////////////////////////

/*!
    \fn CLightwaveObject::loadPTAG(char* type, unsigned long iChunkSize, ifstream* file)
 */
// function for loading the content of the PTAG Chunks
void CLightwaveObject::loadPTAG(char* type, unsigned long iChunkSize, ifstream* file)
{

    // this function is for differ between the types of PTAG Chunks
    vector<ptag>        *target;
    int                 iPolygon;
    unsigned short int  iTag;
    unsigned int        iActualTag = 70000;
    unsigned short int  iIndex, iIndex2;
    ptag*               this_ptag;

// 	if (!strcmp(type,"SURF")) target = &vecPTAG_Surface;		// use the surface vector
// 	if (!strcmp(type,"COLR")) target = &vecPTAG_Color;		// use the color vector

    if (!strcmp(type,"SURF")) {

// 	cout << "SURF" << endl;

        target = &vecPTAG_Surface;

        while (iChunkSize > 0) {

// 	cout << type << ": " <<target->size() << " \n" << flush;

            //read the polygon index
            file->read((char*)&iIndex,sizeof(iIndex));
            iIndex = littleEndian2BigEndian(iIndex);

            // check if four byte form
            if ((iIndex & 0xff00) == 0xff00) {
                file->read((char*)&iIndex2,sizeof(iIndex2));
                iIndex2 = littleEndian2BigEndian(iIndex2);

                iIndex &= 0x00ff;

                iPolygon = (iIndex << 16) | iIndex2;

                iChunkSize -= 4;            // 4 bytes
            }
            //2 byte form
            else {
                iPolygon = iIndex;

                iChunkSize -= 2;            // 2 bytes
            }


            //read the tag index (color or surface)
            file->read((char*)&iTag,sizeof(iTag));
            iTag = littleEndian2BigEndian(iTag);
            iChunkSize -= 2;                     // 2 bytes for the iTag


// 			cout << iPolygon << " - "<< iTag << "\n" << flush;

// 			int tmp = (int)iTag;
            if (iTag == iActualTag) {

                // add this polygon to the actual tag
                this_ptag->vecPolygons.push_back(iPolygon);

            }
            else {
                // update actual tag number
                iActualTag = iTag;
                this_ptag = checkIfSurfaceExists(iTag);

                if (this_ptag == NULL) {		
                    // save the contents of the ptag
                    ptag new_ptag;
                    new_ptag.iTag = iTag;
                    new_ptag.vecPolygons.push_back(iPolygon);
                    target->push_back(new_ptag);
                    this_ptag = &((*target)[target->size() -1]);
                }
                else {

                    this_ptag->vecPolygons.push_back(iPolygon);

                }

            }

    }

    // sort the SURF-PTAG vector
    this->sortPtag(&vecPTAG_Surface, 0, vecPTAG_Surface.size()-1);

// 		cout << " ****************************** " <<endl;
// 		for (int i = 0; i < vecPTAG_Surface.size(); i++)  {
// 			cout << "iTag: " << vecPTAG_Surface[i].iTag << "Size: " << vecPTAG_Surface[i].vecPolygons.size() << endl;
// 		}
// 		cout << " ****************************** " <<endl;

    }
    else file->ignore(iChunkSize);

}

/*!
    \fn CLightwaveObject::loadSurface(unsigned long ChunkSize, ifstream* file)
 */
/**
 * 
 * @param ChunkSize 
 * @param file 
 */
// function for loading the content of the SURF Chunks
void CLightwaveObject::loadSurface(unsigned long ChunkSize, ifstream* file)
{

    this->vecSurfaces.push_back(CSurface());			// add a new element to the surface array


    char*		name = new char[50*sizeof(char)];		// name of the surface
    char*		parent = new char[50*sizeof(char)];		// name of the parent surface

    unsigned short int iSubChunkSize = 0;				// Chunk Size of one subchunk (2 byte integer)
    char cStr[5];							// temporary variable for the Chunk-Type (4 byte character)
    cStr[4] = '\0';

    // get the name of the material
    file->getline (name, 50, '\0' );
    // get the parents name
    file->getline (parent, 50, '\0' );

    cout << "         Surface name: " << name << endl;
// 	cout << parent << endl << flush;

    strcpy(this->vecSurfaces[this->vecSurfaces.size() -1].getName(), name);
    strcpy(this->vecSurfaces[this->vecSurfaces.size() -1].getParentName(), parent);

    ChunkSize -= ( strlen(name) + strlen(parent) + 2 );

    while (file->get() == '\0') ChunkSize--;
    file->unget();

    while ((ChunkSize > 0) && file->good()) {

        unsigned long tmp;

        // read the subchunk name
        file->read(cStr,4);

        // get the Chunk Size of this Subchunk
        file->read((char*)&iSubChunkSize,sizeof(iSubChunkSize));
        iSubChunkSize = littleEndian2BigEndian(iSubChunkSize);

//  		  cout << "Subchunk: " << cStr << " - Size: " << iSubChunkSize << endl;

        // save the color of this material
        if (strcmp("COLR",cStr) == 0) {

            float		*p;
            float		r,g,b;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            p = (float*)&tmp;
            r = *p;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            p = (float*)&tmp;
            g = *p;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            p = (float*)&tmp;
            b = *p;

            vecSurfaces[vecSurfaces.size() -1].setColor( r, g, b );

        }

        // save the luminosity of this material
        else if (strcmp("LUMI",cStr) == 0) {

            float*	fTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            fTemp = (float*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setLuminosity( *fTemp );

        }

        // save the diffuse reflection of this material
        else if (strcmp("DIFF",cStr) == 0) {

            float*	fTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            fTemp = (float*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setDiffuseMaterial( *fTemp );

        }

        // save the specular reflection of this material
        else if (strcmp("SPEC",cStr) == 0) {

            float*	fTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            fTemp = (float*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setSpecularMaterial( *fTemp );

        }

        // save the reflection value of this material
        else if (strcmp("REFL",cStr) == 0) {

            float*	fTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            fTemp = (float*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setReflection( *fTemp );

        }

        // save the transparancy of this material
        else if (strcmp("TRAN",cStr) == 0) {

            float*	fTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            fTemp = (float*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setTransparency( *fTemp );

        }

        // save the translucency of this material
        else if (strcmp("TRNL",cStr) == 0) {

            float*	fTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            fTemp = (float*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setTranslucency( *fTemp );

        }

        // save the glossiness of this material
        else if (strcmp("GLOS",cStr) == 0) {

            int*	iTemp;

            file->read((char*)&tmp,sizeof(tmp));
            tmp = littleEndian2BigEndian(tmp);
            iTemp = (int*)&tmp;

            this->vecSurfaces[this->vecSurfaces.size() -1].setGlossiness( *iTemp );

        }
        // other information of the material is not needed
        else {
            file->ignore(iSubChunkSize -2);
        }

        // the last two byte of the subchunk is a reference to an envelope
        // in fact I do not use envelopes, I don't need this
        /// @TODO if envelopes are implemented don't forget about the SURF Chunk
        file->ignore(2);

        // update the ChunkSize
        ChunkSize -= (4 + 2 + iSubChunkSize);	// (4 byte subchunkname + 2 byte subchunk chunksize + size of the subchunk content)

    }

    delete name;
    delete parent;
	
}

/*!
    \fn CLightwaveObject::checkIfSurfaceExists(unsingned short int surface)
 */
// checks if a surface index allready exists
ptag* CLightwaveObject::checkIfSurfaceExists(unsigned short int surface)
{
    for (unsigned int i = 0; i < this->vecPTAG_Surface.size(); i++) {

        if (vecPTAG_Surface[i].iTag == surface)
            return &(vecPTAG_Surface[i]);

    }

    return NULL;
}


/*!
    \fn CLightwaveObject::loadSurfaceNames(ifstream* file, unsigned long iChunkSize)
 */
// load the surface names from the TAGS field
void CLightwaveObject::loadSurfaceNames(ifstream* file, unsigned long iChunkSize)
{

    // read the data of the TAGS chunk
    while (iChunkSize > 0) {

        char*   name = new char[50*sizeof(char)];           // name of the surface

        // get the name of the material
        file->getline (name, 50, '\0' );

        if (strlen(name) > 0) {
            vecSurfNames.push_back(name);
        }

        iChunkSize -= (strlen(name)+1);


    }

}

/*!
    \fn CLightwaveObject::loadVeritces(ifstream* file, unsigned long iChunkSize
 */
// load the vertex data
void CLightwaveObject::loadVertices(ifstream* file, unsigned long iChunkSize)
{

    unsigned long tmp = 0;
    float* number;

    this->iVertexCount = (int)iChunkSize / 12;

    // fill the vertex structure vector		
    for (int i = 0; i < (int)iChunkSize / 4; i += 3) {

        // add a new element to the Vertex-Vector
        this->vecVertices.push_back(CVertex3D());

        // get the X coordinate og this vertex
        file->read((char*)&tmp,sizeof(unsigned long));
        tmp = littleEndian2BigEndian(tmp);
        number = (float*)&tmp;
        this->vecVertices[this->vecVertices.size() -1].setX( -*number );

        // get the Y coordinate og this vertex
        file->read((char*)&tmp,sizeof(unsigned long));
        tmp = littleEndian2BigEndian(tmp);
        number = (float*)&tmp;
        this->vecVertices[this->vecVertices.size() -1].setY( *number );

        // get the Z coordinate og this vertex
        file->read((char*)&tmp,sizeof(unsigned long));
        tmp = littleEndian2BigEndian(tmp);
        number = (float*)&tmp;
        this->vecVertices[this->vecVertices.size() -1].setZ( *number );
    }

}

/*!
    \fn CLightwaveObject::loadBoundingBox(ifstream* file, unsigned long iChunkSize)
 */
// read the corners of the bounding box
void CLightwaveObject::loadBoundingBox(ifstream* file, unsigned long iChunkSize)
{

    unsigned long tmp = 0;
    float* number;

    // save the upper and lower corner of the BBox
    file->read((char*)&tmp,sizeof(tmp));
    tmp = littleEndian2BigEndian(tmp);
    number = (float*)&tmp;
    this->vBB_lowerCorner.setX( *number );

    file->read((char*)&tmp,sizeof(tmp));
    tmp = littleEndian2BigEndian(tmp);
    number = (float*)&tmp;
    this->vBB_lowerCorner.setY( *number );

    file->read((char*)&tmp,sizeof(tmp));
    tmp = littleEndian2BigEndian(tmp);
    number = (float*)&tmp;
    this->vBB_lowerCorner.setZ( *number );

    file->read((char*)&tmp,sizeof(tmp));
    tmp = littleEndian2BigEndian(tmp);
    number = (float*)&tmp;
    this->vBB_upperCorner.setX( *number );

    file->read((char*)&tmp,sizeof(tmp));
    tmp = littleEndian2BigEndian(tmp);
    number = (float*)&tmp;
    this->vBB_upperCorner.setY( *number );

    file->read((char*)&tmp,sizeof(tmp));
    tmp = littleEndian2BigEndian(tmp);
    number = (float*)&tmp;
    this->vBB_upperCorner.setZ( *number );

}


/*!
    \fn CLightwaveObject::loadPolygons(ifstream* file, unsigned long iChunkSize)
 */
void CLightwaveObject::loadPolygons(ifstream* file, unsigned long iChunkSize)
{
    unsigned int i = 0;
    unsigned short int iNumPolygons = 0;                // number of polygons from the object
    unsigned short int iFaceVertexCount = 0;            // variable for the number of vertices of one polygon
    char cStr[5];
    cStr[4] = '\0'; 

    // type of POLS
    file->read(cStr,4);
    iChunkSize -= 4;

    if (!strcmp(cStr, "FACE")) {
        do {	
            // get the vertex count of the polygon
            file->read((char*)&iFaceVertexCount,sizeof(iFaceVertexCount));
            iFaceVertexCount = littleEndian2BigEndian(iFaceVertexCount);
            i += 2;                                             // 2 bytes for the vertex count

//             unsigned short int tmp = iFaceVertexCount;

            iFaceVertexCount = iFaceVertexCount & 1023;         // mask out highest 6 bit - first 6 bit are flags

            // increase Number of Polygons
            iNumPolygons++;

            // add a new element to the Faces-Vector
            this->cpFaces.push_back(CPolygon(&vecVertices));

            // set number of vertices for this Polygon
            this->cpFaces[this->cpFaces.size() -1].setNumVertices( iFaceVertexCount );

            // get the vertex-indices of this polygon
            for (int j = 0; j < iFaceVertexCount; j++) {
                unsigned short int iIndex;

                file->read((char*)&iIndex,sizeof(iIndex));
                iIndex = littleEndian2BigEndian(iIndex);

                // check if four byte form
                if ((iIndex & 0xff00) == 0xff00) {
                    unsigned short int iIndex2;
                    file->read((char*)&iIndex2,sizeof(iIndex2));
                    iIndex2 = littleEndian2BigEndian(iIndex2);

                    iIndex &= 0x00ff;

                    this->cpFaces[this->cpFaces.size() -1].addVertexIndex((iIndex << 16) | iIndex2);

                    i += 4;	            // 4 bytes for each vertex
                }
                // 2 byte form
                else {
                    this->cpFaces[this->cpFaces.size() -1].addVertexIndex(iIndex);

                    i += 2;                 // 2 bytes for each vertex
                }


            }

        } while (i < iChunkSize);

        this->iFaceCount = iNumPolygons;
    }
    else {
        file->ignore(iChunkSize);
    }

}

/*!
    \fn CLightwaveObject::sortPtag(vector<ptag> *ptag, int start, int end)
 */
// uses the QuickSort algorithm to sort the PTAG-vector ascending
void CLightwaveObject::sortPtag(vector<ptag> *ptag, int start, int end)
{

    int i,j,v,tmp;

    if (end > start) {

        v = (*ptag)[end].iTag; i = start-1; j = end;

        for(;;) {

            while((*ptag)[++i].iTag < v);
            while((*ptag)[--j].iTag > v);
            if (i >= j) break;

            tmp = (*ptag)[i].iTag;
            (*ptag)[i].vecPolygons.swap((*ptag)[j].vecPolygons);
            (*ptag)[i].iTag = (*ptag)[j].iTag;
            (*ptag)[j].iTag = tmp;

        }

        tmp = (*ptag)[i].iTag;
        (*ptag)[i].vecPolygons.swap((*ptag)[end].vecPolygons);
        (*ptag)[i].iTag = (*ptag)[end].iTag;
        (*ptag)[end].iTag = tmp;

        sortPtag( ptag, start, i-1 );
        sortPtag( ptag, i+1, end );

    }

}


///////////////////////////////////////////////////////////////////////////////////
//	Public functions for data access
///////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CLightwaveObject::getSurfacePolygons(int iSurface)
 */
// returns a vector of polygons inherent to the surface
vector<int>* CLightwaveObject::getSurfacePolygons(char* cSurfName)
{
    vector<int>	*polygons;
    unsigned int pos;

    for (pos=0; pos < vecSurfNames.size(); pos++) {
        if (!strcmp(vecSurfNames.at(pos), cSurfName)) {

            for (unsigned int j = 0; j < vecSurfNames.size(); j++) {
                if (vecPTAG_Surface[j].iTag == pos) {
                    polygons = &(vecPTAG_Surface[j].vecPolygons);
                    return polygons;
                }
            }
            return NULL;
        }
    }

    return NULL;
	
}


/*!
    \fn CLightwaveObject::getSurfaceByName(char* name)
 */
CSurface* CLightwaveObject::getSurfaceByName(char* name)
{

    for (unsigned int pos = 0; pos < vecSurfaces.size(); pos++) {

        if (!strcmp(vecSurfaces[pos].getName(), name))
            return (&(vecSurfaces[pos]));

    }

    cout << name << " - Warning: surface with this name does not exist!" << endl;

    return NULL;
	
}


int CLightwaveObject::getFaceCount() const
{
  return iFaceCount;
}


int CLightwaveObject::getVertexCount() const
{
  return iVertexCount;
}


CVertex3D CLightwaveObject::getBBoxLowerCorner() const
{
  return vBB_lowerCorner;
}


CVertex3D CLightwaveObject::getBBoxUpperCorner() const
{
  return vBB_upperCorner;
}

CPolygon CLightwaveObject::getPolygonAtPos( int pos ) const
{
  return cpFaces.at( pos );
}

CVertex3D CLightwaveObject::getVertexAtPos(  int pos ) const
{
  return vecVertices.at( pos );
}
