/***************************************************************************
 *   Copyright (C) 2006 by Holger Gerth                                    *
 *   holger.gerth@informatik.tu-chemnitz.de                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "perfectspatialhash.h"

CPerfectSpatialHash::CPerfectSpatialHash()
{
    this->accessTimeRandomAccess = 0;
    this->accessAverageTimeRandomAccess = 0;
    this->domainBit.init( 0,0,0, false);
}

CPerfectSpatialHash::CPerfectSpatialHash( CVoxelObject3D *vo )
{
    this->accessTimeRandomAccess = 0;
    this->accessAverageTimeRandomAccess = 0;

    this->createPerfectHashFromVoxelobject( vo );
}

CPerfectSpatialHash::~CPerfectSpatialHash()
{
}

bool CPerfectSpatialHash::createPerfectHashFromVoxelobject( CVoxelObject3D *vo ) {
    int                             m = 0;                      // size of the hash table
    int                             r = 0;                      // size of the offset table
    int                             numVoxels = 0;              // number of defined voxels in the voxel object
    int                             x,y,z,i;                    // indices for traversing the coordinates
    CArray3D<pointList*>            tmpTable;                   // temporary array for hash creation
    pointList                       *plNewElement, *p;          // for hash creation
    CPoint3D                        hashPos, offsetPos;         // position in the hash tables
    bool                            notPerfectHashed = false;   // stop condition for hash creation
    coordinateIndex*                indexArray;                 // index array to sort the tmpTable elements

    this->minRange = vo->m_MinRange;
    this->maxRange = vo->m_MaxRange;

    this->minBBox = CPoint3D(0,0,0);
    this->maxBBox = CPoint3D(vo->m_ScopeX,vo->m_ScopeY,vo->m_ScopeZ);

    // get the number of defined voxels for calculating the hash table size and set the domain bit
    this->domainBit.free();
    this->domainBit.init( vo->m_ScopeX, vo->m_ScopeY, vo->m_ScopeZ, 0 );
    for ( z = 0; z < vo->m_ScopeZ; z++ ) {
        for ( y = 0; y < vo->m_ScopeY; y++ ) {
            for ( x = 0; x < vo->m_ScopeX; x++ ) {
                if ( vo->m_CubeArray(x,y,z).flag == CT_BORDER ) {
                    numVoxels++;
                    this->domainBit(x,y,z) = 1;
                }
            }
        }
    }

    // set hash table size
    m = (int) ceil( pow((double)numVoxels,1.0/3.0) );

    // set offset table start size
    r = (int) ceil( pow(( (1.0 / (2.0*3.0))) * (double)numVoxels, 1.0/3.0 ));

    // create the hash table
    do {

        // free all tables
        this->SHData.free();
        this->SOffset.free();
        tmpTable.free();

        // init hash table
        this->SHData.init( m, m, m, -65000.0 );

        // init offset table
        this->SOffset.init( r, r, r, CPoint3D() );

        // init tmp
        tmpTable.init( r, r, r, NULL );

        // fill the temp table
        for ( z = 0; z < vo->m_ScopeZ; z++ ) {
            for ( y = 0; y < vo->m_ScopeY; y++ ) {
                for ( x = 0; x < vo->m_ScopeX; x++ ) {
                    if ( vo->m_CubeArray(x,y,z).flag == CT_BORDER ) {
                        hashPos = this->computeHash( CPoint3D(x,y,z), m );
                        offsetPos = this->computeHash( CPoint3D(x,y,z), r );

                        plNewElement = new pointList;
                        plNewElement->point = hashPos;
                        plNewElement->next = NULL;
                        plNewElement->listSize = 1;
                        plNewElement->value = 1.0;

                        if ( tmpTable(offsetPos.getX(), offsetPos.getY(), offsetPos.getZ()) == NULL ) {
                            tmpTable(offsetPos.getX(), offsetPos.getY(), offsetPos.getZ()) = plNewElement;
                        }
                        else {
                            p = tmpTable(offsetPos.getX(), offsetPos.getY(), offsetPos.getZ());

                            p->listSize++;

                            while ( p->next != NULL )
                                p = p->next;


                            p->next = plNewElement;
                        }
                    }
                }
            }
        }

        // we need at first the elements with the most points in the tmpTable
        // this is an index array which is sorted for this purpose
        indexArray = new coordinateIndex[r*r*r];
        for ( z = 0; z < r; z++ ) {
            for ( y = 0; y < r; y++ ) {
                for ( x = 0; x < r; x++ ) {
                    if ( tmpTable(x,y,z) == NULL ) {
                        indexArray[x + r*y + r*r*z].value = 0;
                    }
                    else {
                        indexArray[x + r*y + r*r*z].value = tmpTable(x,y,z)->listSize;
                        indexArray[x + r*y + r*r*z].coordinate = CPoint3D(x,y,z);
                    }
                }
            }
        }

        // sort the array
        this->quicksort( indexArray, 0, (r*r*r)-1);

        // now we fill the hash table
        i = 0;
        while ( (indexArray[i].value > 0) && (i < (r*r*r))  ) {

            // get the list with the hash coordinates from the temp table
            p = tmpTable(indexArray[i].coordinate.getX(), indexArray[i].coordinate.getY(), indexArray[i].coordinate.getZ());

            // search for free positions
            CPoint3D point = CPoint3D(0,0,0);
            bool posIsFree = this->checkFreePlaces( p, point );

            int z1 = p->point.getZ() -1;
            int z2 = p->point.getZ() +1;
            int y1 = p->point.getY() -1;
            int y2 = p->point.getY() +1;
            int x1 = p->point.getX() -1;
            int x2 = p->point.getX() +1;
            while ( posIsFree == false ) {
                for ( z = z1; z <= z2; z++ ) {
                    if ( (z >= 0) && (z < r) ) {

                        if ( (z == z1) || (z == z2) ) {
                            for ( y = y1; y <= y2; y++ ) {
                                if ( (y >= 0) && (y < r) ) {
                                    for ( x = x1; x <= x2; x++ ) {

                                        if ((x >= 0) && (y >= 0) && (z >= 0) && (x < r) && (y < r) && (z < r)) {
                                            if ( (z > z1) && (z < z2) ) {
                                                if ( (x == x1) || (x == x2) || (y == y1) || (y == y2) ) {
                                                    point = CPoint3D(x - p->point.getX(), y - p->point.getY(), z - p->point.getZ() );
                                                    posIsFree = this->checkFreePlaces( p, point );
                                                }
                                            }
                                            else {
                                                point = CPoint3D(x - p->point.getX(), y - p->point.getY(), z - p->point.getZ() );
                                                posIsFree = this->checkFreePlaces( p, point );
                                            }
                                        }

                                        if ( posIsFree ) break;
                                    }
                                }
                                if ( posIsFree ) break;
                            }
                        }
                        else {
                            for ( y = y1; y <= y2; y++ ) {
                                if ( (y >= 0) && (y < r) ) {
                                    if ( ( y == y1) && (y == y2)) {
                                        for ( x = x1; x <= x2; x++ ) {

                                            if ((x >= 0) && (y >= 0) && (z >= 0) && (x < r) && (y < r) && (z < r)) {
                                                if ( (z > z1) && (z < z2) ) {
                                                    if ( (x == x1) || (x == x2) || (y == y1) || (y == y2) ) {
                                                        point = CPoint3D(x - p->point.getX(), y - p->point.getY(), z - p->point.getZ() );
                                                        posIsFree = this->checkFreePlaces( p, point );
                                                    }
                                                }
                                                else {
                                                    point = CPoint3D(x - p->point.getX(), y - p->point.getY(), z - p->point.getZ() );
                                                    posIsFree = this->checkFreePlaces( p, point );
                                                }
                                            }

                                            if ( posIsFree ) break;
                                        }
                                    }
                                    else {
                                        if (x1 >= 0) {
                                            point = CPoint3D(x1 - p->point.getX(), y - p->point.getY(), z - p->point.getZ() );
                                            posIsFree = this->checkFreePlaces( p, point );
                                        }
                                        if ( (posIsFree == false) && (x2 < r) ) {
                                            point = CPoint3D(x1 - p->point.getX(), y - p->point.getY(), z - p->point.getZ() );
                                            posIsFree = this->checkFreePlaces( p, point );
                                        }
                                    }
                                }
                                if ( posIsFree ) break;
                            }
                        }

                    }
                    if ( posIsFree ) break;
                }
                z1--; z2++; y1--; y2++; x1--; x2++;

                // no more free positions in the hash table
                if ( (posIsFree == false) && (x1 < 0) && (y1 < 0) && (z1 < 0) && (x2 >= r) && (y2 >= r) && (z2 >= r) ) {
                    break;
                }
            }

            // insert the points if positions in the hash table are free
            if ( posIsFree ) {
                this->SOffset(indexArray[i].coordinate.getX(), indexArray[i].coordinate.getY(), indexArray[i].coordinate.getZ()) = point;
                while ( p != NULL ) {
                    this->SHData(p->point.getX(), p->point.getY(), p->point.getZ()) = p->value;
                    p = p->next;
                }
                notPerfectHashed = false;
            }
            // no more free positions so we need to increase the size of the offset table
            else {
                r++;
                notPerfectHashed = true;

                break;
            }

            i++;
        }

        delete [] indexArray;
    } while ( notPerfectHashed );

    return true;
}

// computes the position in the hash table
CPoint3D CPerfectSpatialHash::computeHash( CPoint3D p, int size ) {

    CPoint3D returnValue;

    returnValue.setX( p.getX() % size );
    returnValue.setY( p.getY() % size );
    returnValue.setZ( p.getZ() % size );

    return returnValue;

}

// random access operator
double CPerfectSpatialHash::randomAccess( int x, int y, int z ) {
    return this->randomAccess( CPoint3D(x,y,z) );
}

// random access operator
double CPerfectSpatialHash::randomAccess( CPoint3D point ) {
    CPoint3D offsetPos = this->computeHash( point, this->SOffset.width() );
    CPoint3D hashPos = this->computeHash( point, this->SHData.width() );
    CPoint3D offset = this->SOffset(offsetPos.getX(), offsetPos.getY(), offsetPos.getZ());

    return this->SHData( offset.getX() + hashPos.getX(), offset.getY() + hashPos.getY(), offset.getZ() + hashPos.getZ() );
}

//////////////////////////////////////////////////////////////////////////////////////////////
//                      Private functions                                                   //
//////////////////////////////////////////////////////////////////////////////////////////////

// sorting algorithm for the index array
void CPerfectSpatialHash::quicksort( coordinateIndex* array, int min, int max ) {

    int links = min, rechts = max, pivot;
    coordinateIndex tmp;

    // two elements
    if ( (max - min) > 0 ) {
        pivot = max;
        while ( links < rechts ) {
            while ( (array[links].value >= array[pivot].value) ) {
                if (links >= pivot) break;
                links++;
            }
            while ( (array[rechts].value <= array[pivot].value) ) {
                if (rechts <= pivot) break;
                rechts--;
            }

            tmp.coordinate = array[links].coordinate;
            tmp.value = array[links].value;
            array[links].coordinate = array[rechts].coordinate;
            array[links].value = array[rechts].value;
            array[rechts].coordinate = tmp.coordinate;
            array[rechts].value = tmp.value;

            if ( links == pivot ) pivot = rechts;
            else if ( rechts == pivot ) pivot = links;
        }

        // sort the two parts recursively
        if ( min < (pivot-1) ) this->quicksort( array, min, pivot-1 );
        if ( max > (pivot+1) ) this->quicksort( array, pivot+1, max );

    }

}

// check if position in the hash table are free
bool CPerfectSpatialHash::checkFreePlaces( pointList* p, CPoint3D point ) {

    if ( p != NULL ) {
        if ( this->SHData(p->point.getX() + point.getX(), p->point.getY() + point.getY(), p->point.getZ() + point.getZ()) > -65000.0 )
            return false;

        if ( p->next != NULL )
            return this->checkFreePlaces( p->next, point );

        return true;
    }
    else return false;
}


//////////////////////////////////////////////////////////////////////////////////////////////
//                      Testing                                                             //
//////////////////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CPerfectSpatialHash::drawCubes()
 */
void CPerfectSpatialHash::drawCubes()
{
    // for comnputing the average access time of the random access operator
    long            numGridPoints = 0, accessTime = 0, totalAccessTime = 0, averageAccessTime = 0;
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)

    if (d == 3) {

        int minX = this->minBBox.getX();
        int maxX = this->maxBBox.getX();
        int minY = this->minBBox.getY();
        int maxY = this->maxBBox.getY();
        int minZ = this->minBBox.getZ();
        int maxZ = this->maxBBox.getZ();

        float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
        float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
        float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

        for (int z = minZ; z < maxZ; z++) {
            for (int y = minY; y < maxY; y++) {
                for (int x = minX; x < maxX; x++) {

                    gettimeofday( &tp1, NULL );
                    if (this->domainBit(x,y,z) == true) {
                    gettimeofday( &tp2, NULL );
                    accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);

                        gettimeofday( &tp1, NULL );
                        double value = this->randomAccess( x, y, z );
                        gettimeofday( &tp2, NULL );

                        accessTime += (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);

                        // draw cubes and cube edges
                        // -------------------------

                        CVertex3D point1( this->minRange.getX() + x * CubeW,
                                          this->minRange.getY() + y * CubeH,
                                          this->minRange.getZ() + z * CubeD );
                        CVertex3D point2( this->minRange.getX() + (x+1) * CubeW,
                                          this->minRange.getY() + (y+1) * CubeH,
                                          this->minRange.getZ() + (z+1) * CubeD);

                        this->drawACube(&point1, &point2, value);

                    }
                    else {
                        gettimeofday( &tp2, NULL );
                        accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);
                    }

                    totalAccessTime += accessTime;
                    numGridPoints++;

                }
            }
        }

    }

    averageAccessTime = totalAccessTime / numGridPoints;

    cout << "Total time while accessing the hash tables was " << totalAccessTime << " micro seconds and the average time";
    cout << " for one element was " << averageAccessTime << " micro seconds." << endl;
    cout << "Last random access done in " << accessTime << " micro seconds." << endl;

    this->accessTimeRandomAccess = (double) accessTime;
    this->accessAverageTimeRandomAccess = (double) totalAccessTime / (double) numGridPoints;
    this->accessTimeSequentialAccess = (double) totalAccessTime;
}


void CPerfectSpatialHash::drawACube(CVertex3D *p1, CVertex3D *p2, double colorValue) {

    // Flaechen zeichnen
    // ----------------
        // Farbe festlegen
        glColor3f(0.5, 1.0, 0.2 * colorValue);

        // Offset fr den Cube (bessere Darstellung der Kanten)
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(1.f, 1.f);

            glBegin(GL_POLYGON);
                glNormal3f(  -1.0,  0.0,  0.0);
                glVertex3f(  p1->getX(), p1->getY(), p1->getZ());
                glVertex3f(  p1->getX(), p1->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d( 1.0,  0.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  -1.0,  0.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0, 1.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, -1.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, 1.0);
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

        glDisable(GL_POLYGON_OFFSET_FILL);


// Kanten zeichnen
// ---------------

        glColor3d(0.2, 0.2, 0.2);

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();


    return;
}

void CPerfectSpatialHash::printDataStructure() {


}

/*!
    \fn CPerfectSpatialHash::getMemorySize()
 */
// returns the size of the data structure in bytes
double CPerfectSpatialHash::getMemorySize()
{
    double returnValue = 0.0;

    returnValue += this->SOffset.height() * this->SOffset.width() * this->SOffset.depth() * sizeof( CPoint3D );
    returnValue += this->SHData.height() * this->SHData.width() * this->SHData.depth() * sizeof( double );
    returnValue += (this->maxBBox.getX() - this->minBBox.getX() ) * (this->maxBBox.getY() - this->minBBox.getY() ) * (this->maxBBox.getZ() - this->minBBox.getZ() ) * sizeof( bool );

    return returnValue;
}
