/***************************************************************************
 *   Copyright (C) 2005 by Holger Gerth                                    *
 *   holger.gerth@informatik.tu-chemnitz.de                                *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "qopenglwidget.h"

CqOpenGLwidget::CqOpenGLwidget(QWidget *parent) : QGLWidget(parent) {
    // constructor
    this->camera = new CoglCamera3D();
    this->myMouse = CMouse();

    // OpenGL status flags
    this->bLighting = false;
    this->bCoordinates = false;
    this->bWireFrame = false;
    this->bColorMaterial = false;
    this->bCullBackFace = false;
    this->bTransparency = false;

    this->bSmfObjectStatus = false;
    this->bVoxelObjectStatus = false;
    this->bHrleObjectStatus = false;
    this->bDtGridObjectStatus = false;
    this->bPerfectHashObjectStatus = false;
    this->bSparseLevelSetStatus = false;

    this->bSmfObjectChanged = false;
    this->bVoxelObjectChanged = false;
    this->bHrleObjectChanged = false;
    this->bDtGridObjectChanged = false;
    this->bPerfectHashObjectChanged = false;
    this->bSparseLevelSetChanged =false;

    this->bDrawVoxelObject = false;
    this->bDrawSmfObject = false;
    this->bDrawHrleRand = false;
    this->bDrawHrleSeq = false;
    this->bDrawDtRand = false;
    this->bDrawDtSeq = false;
    this->bDrawPHRand = false;
    this->bDrawSparseLevelSetRand = false;
    this->bDrawSparseLevelSetSeq = false;

    this->iCameraMode = 0;

    this->myHRLE = NULL;
    this->myDtGrid = NULL;
    this->mySmfObject = NULL;
    this->myVoxelObject = NULL;
    this->myPerfectHash = NULL;
    this->mySparseLevelSet = NULL;

    // start the idle timer
    this->idleTimer = new QTimer (this);
    connect( idleTimer, SIGNAL( timeout() ), SLOT( idleFuncGL() ) );
    this->idleTimer->start(0);
}

CqOpenGLwidget::~CqOpenGLwidget() {

    delete idleTimer;
    delete camera;

    if ( this->myHRLE != NULL ) delete this->myHRLE;
    if ( this->myDtGrid != NULL ) delete this->myDtGrid;
    if ( this->mySmfObject != NULL ) delete this->mySmfObject;
    if ( this->myVoxelObject != NULL ) delete this->myVoxelObject;
    if ( this->myPerfectHash != NULL ) delete this->myPerfectHash;
    if ( this->mySparseLevelSet != NULL ) delete this->mySparseLevelSet;

}

void CqOpenGLwidget::initializeGL() {

    // GL state initializations

    // set background color to black
    glClearColor(0.01, 0.01, 0.1, 1.0);

    // shading model
    glShadeModel(GL_SMOOTH);

    // clear Z-buffer
    glClear(GL_DEPTH_BUFFER_BIT);

    // Z-buffer
    glEnable(GL_DEPTH_TEST);

    // normalize normals
    glEnable(GL_NORMALIZE);

    glFrontFace(GL_CW);

    // initialize viewing values
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // set global ambient light
    float glob_amb[] = {0.2,0.2,0.2};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, glob_amb);

    // LightModel Local Viewer
//      glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
//      glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

    // set mode for color material
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

    // set clipping planes
    glFrustum(
            -0.5,               // left Plane
            +0.5,               // right Plane
            -0.5,               // upper Plane
            +0.5,               // lower Plane
            +1.0,               // near
            +2000);             // far

    // switch to modelview matrix
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // load static gl draw lists
    staticListsGL();

}

void CqOpenGLwidget::paintGL() {
    // draw scene

    // clear color and depth buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //draw wireframe
    if (this->bWireFrame)
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    else
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    // switch to modelview matrix
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    camera->setCameraMode( iCameraMode );
    camera->setCamPos();

///////////////////////////////////////////////////////////////////////////////////////////
//      light definitions
///////////////////////////////////////////////////////////////////////////////////////////

    // Lightsource 0
    float light0_pos[] = {5.0, 0.0, -2.0, 0.0};
    float light0_amb[] = {0.1, 0.1, 0.1};
    float light0_dif[] = {0.5, 0.5, 0.5};
    float light0_spec[] = {0.5, 0.5, 0.5};

    glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
    glLightfv(GL_LIGHT0, GL_AMBIENT, light0_amb);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_dif);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_spec);

    // Lightsource 1
    float light1_pos[] = {0.0, 5.0, 2.0, 0.0};
    float light1_amb[] = {0.0, 0.0, 0.0};
    float light1_dif[] = {0.7, 0.7, 0.7};
    float light1_spec[] = {0.5, 0.5, 0.5};

    glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
    glLightfv(GL_LIGHT1, GL_AMBIENT, light1_amb);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_dif);
    glLightfv(GL_LIGHT1, GL_SPECULAR, light1_spec);

    // Lightsource 2
    float light2_pos[] = {-1.0, 1.0, 1.0, 0.0};
    float light2_amb[] = {0.0, 0.0, 0.0};
    float light2_dif[] = {0.5, 0.5, 0.5};
    float light2_spec[] = {0.5, 0.5, 0.5};

    glLightfv(GL_LIGHT2, GL_POSITION, light2_pos);
    glLightfv(GL_LIGHT2, GL_AMBIENT, light2_amb);
    glLightfv(GL_LIGHT2, GL_DIFFUSE, light2_dif);
    glLightfv(GL_LIGHT2, GL_SPECULAR, light2_spec);

    // Lightsource 3
    float light3_pos[] = {1.0, 1.0, -1.0, 0.0};
    float light3_amb[] = {0.0, 0.0, 0.0};
    float light3_dif[] = {0.5, 0.5, 0.5};
    float light3_spec[] = {0.5, 0.5, 0.5};

    glLightfv(GL_LIGHT3, GL_POSITION, light3_pos);
    glLightfv(GL_LIGHT3, GL_AMBIENT, light3_amb);
    glLightfv(GL_LIGHT3, GL_DIFFUSE, light3_dif);
    glLightfv(GL_LIGHT3, GL_SPECULAR, light3_spec);

///////////////////////////////////////////////////////////////////////////////////////////
//		change OpenGL state
///////////////////////////////////////////////////////////////////////////////////////////	

    // set OpenGL status
        // Lighting
        if (this->bLighting)
            glEnable(GL_LIGHTING);
        else
            glDisable(GL_LIGHTING);

        glEnable(GL_LIGHT0);
        glEnable(GL_LIGHT1);
//         glEnable(GL_LIGHT2);
//         glEnable(GL_LIGHT3);

    // front face to counter clockwise
    glFrontFace(GL_CCW);

    // glColor-Modus
    if (this->bColorMaterial)
        glEnable(GL_COLOR_MATERIAL);
    else
        glDisable(GL_COLOR_MATERIAL);

    // backface Culling
    if (this->bCullBackFace) 
        glEnable(GL_CULL_FACE);
    else 
        glDisable(GL_CULL_FACE);

    // Alpha Blending
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    if (this->bTransparency)
        glEnable(GL_BLEND);
    else
        glDisable(GL_BLEND);

    // draw coordinates
    if (this->bCoordinates)
        glCallList((GLint) COORDINATES);

///////////////////////////////////////////////////////////////////////////////////////////
//		Draw the object
///////////////////////////////////////////////////////////////////////////////////////////

//     CWavingFlag flag;
//     flag.drawWavingFlag();

////////////////
// --> fr Studienarbeit
///////////////

    glPushMatrix();

        glTranslatef(0, 1, 0);

        if ( this->bSmfObjectStatus && this->bDrawSmfObject ) {
            glCallList((GLint) SMFOBJECT);
        }

        if ( this->bVoxelObjectStatus && this->bDrawVoxelObject ) {
            glCallList((GLint) VOXELOBJECT);
        }

        if ( this->bHrleObjectStatus && this->bDrawHrleRand ) {
            glCallList((GLint) HRLE_RANDOM);
        }

        if ( this->bHrleObjectStatus && this->bDrawHrleSeq ) {
            glCallList((GLint) HRLE_SEQUENTIAL);
        }

        if ( this->bDtGridObjectStatus && this->bDrawDtRand ) {
            glCallList((GLint) DT_RANDOM);
        }

        if ( this->bDtGridObjectStatus && this->bDrawDtSeq ) {
            glCallList((GLint) DT_SEQUENTIAL);
        }

        if ( this->bPerfectHashObjectStatus && this->bDrawPHRand ) {
            glCallList((GLint) PH_RANDOM);
        }

        if ( this->bSparseLevelSetStatus && this->bDrawSparseLevelSetRand ) {
            glCallList((GLint) SPARSE_RANDOM);
        }

        if ( this->bSparseLevelSetStatus && this->bDrawSparseLevelSetSeq ) {
            glCallList((GLint) SPARSE_SEQUENTIAL);
        }

    glPopMatrix();

///////////////////////////////////////////////////////////////////////////////////////////
//		end of drawing
///////////////////////////////////////////////////////////////////////////////////////////

    swapBuffers();
    glFlush();

}

/*!
    \fn CqOpenGLwidget::mouseMoveEvent( QMouseEvent *ev )
*/
void CqOpenGLwidget::mouseMoveEvent( QMouseEvent *ev )
{
// mouse movement callback function

    if ( QEvent::MouseMove ) {

        myMouse.setCurX( ev->globalX() );
        myMouse.setCurY( ev->globalY() );

        switch (myMouse.getLastState()) {
            case (Qt::LeftButton):
                myMouse.setMoveX( myMouse.getCurX() - myMouse.getOldX() );
                myMouse.setMoveY( myMouse.getCurY() - myMouse.getOldY() );
                break;
            case (Qt::RightButton):
                myMouse.setRadius( myMouse.getRadius() + myMouse.getCurY() - myMouse.getOldY() );
                myMouse.setMoveY( myMouse.getCurY() - myMouse.getOldY() );
                break;

            default: break;
        }

        if (myMouse.getRadius() < 1.0) myMouse.setRadius( 1.0 );

        camera->updateMouseData( myMouse.getMoveX(), myMouse.getMoveY(), myMouse.getRadius(), myMouse.getLastState());

        myMouse.setOldX( ev->globalX() );
        myMouse.setOldY( ev->globalY() );

        updateGL();
    }
}


/*!
    \fn CqOpenGLwidget::mousePressEvent( QMouseEvent *ev )
*/
void CqOpenGLwidget::mousePressEvent( QMouseEvent *ev )
{
    // if mouse button is pressed

    if (ev->button() != Qt::NoButton) myMouse.setLastState( ev->button() );
    else myMouse.setLastState( -1 );

    myMouse.setOldX( ev->globalX() );
    myMouse.setOldY( ev->globalY() );
}


/*!
    \fn CqOpenGLwidget::mouseReleaseEvent( QMouseEvent *ev )
*/
void CqOpenGLwidget::mouseReleaseEvent( QMouseEvent *ev )
{
    // if mouse button is released

    if (ev->button() == Qt::NoButton)myMouse.setLastState( -1 );

}


void CqOpenGLwidget::resizeGL(int w, int h) {
    // handle resize events

    // set viewport to window size
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);

    updateGL();

}

/*!
    \fn CqOpenGLwidget::staticListsGL()
*/
void CqOpenGLwidget::staticListsGL()
{
    // static lists loaded at startup

    // draw coordinate system
    glNewList((GLint) COORDINATES, GL_COMPILE);
            glPushAttrib(GL_LIGHTING_BIT);
            glPushMatrix();
                    glDisable(GL_LIGHTING);
                    glLineWidth(3);
                    glBegin(GL_LINES);
                            //X axis (red)
                            glColor4f(1.0,0.0,0.0,1.0);
                            glVertex3d(0.0,0.0,0.0);
                            glVertex3d(5.0,0.0,0.0);

                            // Y axis (green)
                            glColor4f(0.0,1.0,0.0,1.0);
                            glVertex3d(0.0,0.0,0.0);
                            glVertex3d(0.0,5.0,0.0);

                            // Z axis (blue)
                            glColor4f(0.0,0.0,1.0,1.0);
                            glVertex3d(0.0,0.0,0.0);
                            glVertex3d(0.0,0.0,5.0);
                    glEnd();
                    glLineWidth(1);
            glPopMatrix();
            glPopAttrib();
    glEndList();

}


/*!
    \fn CqOpenGLwidget::dynListsGL()
*/
//non-static GL-Lists
void CqOpenGLwidget::dynListsGL()
{
    // GL lists which can be changed during execution

    if ( this->bSmfObjectStatus && this->bSmfObjectChanged ) {
        glNewList((GLint) SMFOBJECT, GL_COMPILE);
            glPushMatrix();
                mySmfObject->draw();
            glPopMatrix();
        glEndList();
    }

    if ( this->bVoxelObjectStatus && this->bVoxelObjectChanged ) {
        glNewList((GLint) VOXELOBJECT, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 0, 1);
                myVoxelObject->Draw();
//                 myVoxelObject->drawCubesUsingSequentialAccess();
            glPopMatrix();
        glEndList();
    }

    if ( this->bHrleObjectStatus && this->bHrleObjectChanged ) {
        glNewList((GLint) HRLE_RANDOM, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 0, -1);
                myHRLE->drawCubes();
            glPopMatrix();
        glEndList();
    }

    if ( this->bHrleObjectStatus && this->bHrleObjectChanged ) {
        glNewList((GLint) HRLE_SEQUENTIAL, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 0, -1);
                myHRLE->drawCubesUsingSequentialAccess();
            glPopMatrix();
        glEndList();
    }

    if ( this->bDtGridObjectStatus && this->bDtGridObjectChanged ) {
        glNewList((GLint) DT_RANDOM, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 1, -1);
                myDtGrid->drawCubes();
            glPopMatrix();
        glEndList();
    }

    if ( this->bDtGridObjectStatus && this->bDtGridObjectChanged ) {
        glNewList((GLint) DT_SEQUENTIAL, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 1, -1);
                myDtGrid->drawCubesUsingSequentialAccess();
            glPopMatrix();
        glEndList();
    }

    if ( this->bPerfectHashObjectStatus && this->bPerfectHashObjectChanged ) {
        glNewList((GLint) PH_RANDOM, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 1, 0);
                myPerfectHash->drawCubes();
            glPopMatrix();
        glEndList();
    }

    if ( this->bSparseLevelSetStatus && this->bSparseLevelSetChanged ) {
        glNewList((GLint) SPARSE_RANDOM, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 1, 1);
                mySparseLevelSet->drawCubes();
            glPopMatrix();
        glEndList();
    }

    if ( this->bSparseLevelSetStatus && this->bSparseLevelSetChanged ) {
        glNewList((GLint) SPARSE_SEQUENTIAL, GL_COMPILE);
            glPushMatrix();
                glTranslatef(0, 1, 1);
                mySparseLevelSet->drawCubesUsingSequentialAccess();
            glPopMatrix();
        glEndList();
    }

    this->bSmfObjectChanged = false;
    this->bVoxelObjectChanged = false;
    this->bHrleObjectChanged = false;
    this->bDtGridObjectChanged = false;
    this->bPerfectHashObjectChanged = false;
    this->bSparseLevelSetChanged = false;

    updateGL();

}

/*!
    \fn CqOpenGLwidget::drawSurface(char* surface, LightwsaveObject* object)
*/
// draw a surface
void CqOpenGLwidget::drawSurface(char* surface, CLightwaveObject* object)
{
    // get the polygons of this surface
    vector<int>* polygons = object->getSurfacePolygons( surface );              // returns a vector of indices

    if (polygons != NULL) {
        // draw the polygons
        for (int j = 0; j < polygons->size(); j++) {

            CPolygon Face = object->getPolygonAtPos( polygons->at(j) );         // get the polygon

            CVector3D vNormal = Face.getNormalVector();

            unsigned short int iVertexCount = Face.getNumVertices();

            if (iVertexCount >= 3) {
                glBegin(GL_POLYGON);

                    glNormal3f(vNormal.getX(), vNormal.getY(), vNormal.getZ());

                for (int p = 0; p < iVertexCount; p++) {
                    int pos = Face.getVertexIndexAtPos( p );
                    CVertex3D tmpVertex = object->getVertexAtPos( pos );
                    glVertex3f(	tmpVertex.getX(),
                                tmpVertex.getY(),
                                tmpVertex.getZ()
                    );
                }
                glEnd();
            }
        }
    }
    else cout << "Surface " << surface << " does not exist!" << endl;
}


/*!
    \fn CqOpenGLwidget::setColor(char* surface, LightwaveObject* object, float gamma = 1.0)
*/
// set the color of a surface
void CqOpenGLwidget::setColor(char* surface, CLightwaveObject* object, float gamma, bool on)
{
    CSurface* sf = object->getSurfaceByName( surface );
    CColor4f surfaceColor = sf->getColor();

    if (sf != NULL) {

        if (on) {

            float mat_amb[] = {	    sf->getAmbientMaterial() * surfaceColor.getRed() * gamma,
                                    sf->getAmbientMaterial() * surfaceColor.getGreen() * gamma,
                                    sf->getAmbientMaterial() * surfaceColor.getBlue() * gamma,
                                    1};
            float mat_diff[] = {    sf->getDiffuseMaterial() * surfaceColor.getRed() * gamma,
                                    sf->getDiffuseMaterial() * surfaceColor.getGreen() * gamma,
                                    sf->getDiffuseMaterial() * surfaceColor.getBlue() * gamma,
                                    1};
            float mat_spec[] = {    sf->getSpecularMaterial() * surfaceColor.getRed(),
                                    sf->getSpecularMaterial() * surfaceColor.getGreen(),
                                    sf->getSpecularMaterial() * surfaceColor.getBlue(),
                                    1};

            float mat_emm[] = {     surfaceColor.getRed()  * gamma,
                                    surfaceColor.getGreen() * gamma,
                                    surfaceColor.getBlue() * gamma,
                                    1};

            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_amb);
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diff);
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_spec);

            glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mat_emm);
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (GLfloat)(sf->getGlossiness() * 128/1024));

            // set color for color mode
            glColor4f(  surfaceColor.getRed() * gamma,
                        surfaceColor.getGreen() * gamma,
                        surfaceColor.getBlue() * gamma,
                        1);

        }
        else {

            float mat_amb[] = {	    sf->getAmbientMaterial() * surfaceColor.getRed() * gamma,
                                    sf->getAmbientMaterial() * surfaceColor.getGreen() * gamma,
                                    sf->getAmbientMaterial() * surfaceColor.getBlue() * gamma,
                                    sf->getTransparency()};
            float mat_diff[] = {    sf->getDiffuseMaterial() * surfaceColor.getRed() * gamma,
                                    sf->getDiffuseMaterial() * surfaceColor.getGreen() * gamma,
                                    sf->getDiffuseMaterial() * surfaceColor.getBlue() * gamma,
                                    sf->getTransparency()};
            float mat_spec[] = {    sf->getSpecularMaterial() * surfaceColor.getRed(),
                                    sf->getSpecularMaterial() * surfaceColor.getGreen(),
                                    sf->getSpecularMaterial() * surfaceColor.getBlue(),
                                    sf->getTransparency()};

            float mat_emm[] = {     sf->getLuminosity() * surfaceColor.getRed() * gamma,
                                    sf->getLuminosity() * surfaceColor.getGreen() * gamma,
                                    sf->getLuminosity() * surfaceColor.getBlue() * gamma,
                                    sf->getTransparency()};

            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_amb);
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diff);
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_spec);

            glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mat_emm);
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (GLfloat)(sf->getGlossiness() * 128/1024));

            // set color for color mode
            glColor4f(  surfaceColor.getRed() * gamma,
                        surfaceColor.getGreen() * gamma,
                        surfaceColor.getBlue() * gamma,
                        sf->getTransparency());

        }

    }
}


/*!
    \fn CqOpenGLwidget::idleFuncGL()
*/
// idle function
void CqOpenGLwidget::idleFuncGL()
{

        updateGL();
}

/*!
    \fn CqOpenGLwidget::loadVoxelObject( int numCubes )
 */
// create a voxel object from an triangulated object
bool CqOpenGLwidget::loadVoxelObject( int numCubes )
{

    if ( this->bVoxelObjectStatus )
        delete this->myVoxelObject;

    this->myVoxelObject = new CVoxelObject3D();
    this->myVoxelObject->TransformTriangleMeshToVoxelSet( this->mySmfObject, numCubes );
    this->bVoxelObjectStatus = true;
    this->myVoxelObject->countDefinedVoxels();

    this->bVoxelObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::loadSmfObject( char* filename )
 */
// load triangulated object from a .smf file
bool CqOpenGLwidget::loadSmfObject( const char* filename )
{
    if ( this->bSmfObjectStatus )
        delete this->mySmfObject;

    this->mySmfObject = new CsmfObject( filename );
    this->bSmfObjectStatus = true;

    this->bSmfObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::loadHrleObjectFromVoxelObject()
 */
// create a HRLE object from a voxel object
bool CqOpenGLwidget::loadHrleObjectFromVoxelObject()
{
    if ( this->bHrleObjectStatus )
        delete this->myHRLE;

    this->myHRLE = new CHierarchicalRLE( 3 );
    this->myHRLE->convert_to_hrle( myVoxelObject );
    this->bHrleObjectStatus = true;

//     cout << "oooooooooooooooooooooooooooooooooooooooooooooo H-RLE object created oooooooooooooooooooooooooooooooooooooo" << endl;
//     this->myHRLE->printDataStructure();
//     cout << "oooooooooooooooooooooooooooooooooooooooooooooo oooooooooooooooooooo oooooooooooooooooooooooooooooooooooooo" << endl;

    this->bHrleObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::loadDtGridFromVoxelObject()
 */
bool CqOpenGLwidget::loadDtGridFromVoxelObject()
{
    if ( this->bDtGridObjectStatus )
        delete this->myDtGrid;

    myDtGrid = new CDtGrid( 3 );
    myDtGrid->convert_to_dtgrid( myVoxelObject );
    this->bDtGridObjectStatus = true;

//     cout << "oooooooooooooooooooooooooooooooooooooooooooooooo DT Grid created ooooooooooooooooooooooooooooooooooooooooo" << endl;
//     myDtGrid->printDataStructure();
//     cout << "ooooooooooooooooooooooooooooooooooooooooooooooooo oooooooooooooo ooooooooooooooooooooooooooooooooooooooooo" << endl;


    this->bDtGridObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::dilateHRLEObject( int dilationPoints )
 */
bool CqOpenGLwidget::dilateHRLEObject( int dilationPoints ) {

    CHierarchicalRLE* tmpHRLE = myHRLE->getDilatedHRLE( dilationPoints );

// cout << "oooooooooooooooooooooooooooooooooooooooooooooo after Dilation oooooooooooooooooooooooooooooooooooooo" << endl;
// tmpHRLE->printDataStructure();
// cout << "oooooooooooooooooooooooooooooooooooooooooooooo oooooooooooooo oooooooooooooooooooooooooooooooooooooo" << endl;

    delete myHRLE;
    myHRLE = tmpHRLE;

    this->bHrleObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::removeFarPoints( double distance )
 */
bool CqOpenGLwidget::removeFarPoints( double distance ) {

    CHierarchicalRLE* tmpHRLE = myHRLE->removeFarPoints( distance );

// cout << "oooooooooooooooooooooooooooooooooooooooooooooo after removing oooooooooooooooooooooooooooooooooooooo" << endl;
// tmpHRLE->printDataStructure();
// cout << "oooooooooooooooooooooooooooooooooooooooooooooo oooooooooooooo oooooooooooooooooooooooooooooooooooooo" << endl;

    delete myHRLE;
    myHRLE = tmpHRLE;

    this->bHrleObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::loadPerfectHashFromVoxelObject()
 */
bool CqOpenGLwidget::loadPerfectHashFromVoxelObject()
{
    if ( this->bPerfectHashObjectStatus )
        delete this->myPerfectHash;

    myPerfectHash = new CPerfectSpatialHash( myVoxelObject );
    this->bPerfectHashObjectStatus = true;

//     cout << "oooooooooooooooooooooooooooooooooooooooooooooo Perfect Hash created ooooooooooooooooooooooooooooooooooooooooo" << endl;
//     myPerfectHash->printDataStructure();
//     cout << "ooooooooooooooooooooooooooooooooooooooooooooooooo oooooooooooooo ooooooooooooooooooooooooooooooooooooooooo" << endl;

    this->bPerfectHashObjectChanged = true;

    this->dynListsGL();

    return true;
}

/*!
    \fn CqOpenGLwidget::loadSparseLevelSetFromVoxelObject()
 */
bool CqOpenGLwidget::loadSparseLevelSetFromVoxelObject()
{
    if ( this->bSparseLevelSetStatus )
        delete this->mySparseLevelSet;

    mySparseLevelSet = new CRLESparseLevelSet( myVoxelObject );
    this->bSparseLevelSetStatus = true;

//     cout << "ooooooooooooooooooooooooooooooooooooooooo RLE Sparse Level Set created ooooooooooooooooooooooooooooooooooooooooo" << endl;
//     mySparseLevelSet->printDataStructure();
//     cout << "ooooooooooooooooooooooooooooooooooooooooooooooooo oooooooooooooo ooooooooooooooooooooooooooooooooooooooooooooooo" << endl;

    this->bSparseLevelSetChanged = true;

    this->dynListsGL();

    return true;
}

// for testing
void CqOpenGLwidget::createTestingHRLE() {

    run_code newRun;

    myHRLE = new CHierarchicalRLE( 3 );

    // dimension 3
    myHRLE->minRange = CVertex3D(0,0,0);
    myHRLE->maxRange = CVertex3D(5,5,5);
    myHRLE->min = 0;
    myHRLE->max = 8;

    newRun.type = RT_POSITIVE;
    myHRLE->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 0;
    myHRLE->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->run_types.push_back( newRun );

    myHRLE->run_breaks.push_back( 1 );
    myHRLE->run_breaks.push_back( 7 );

    myHRLE->start_indices.push_back( 0 );

    // dimension 2
    myHRLE->lower_hrle->min = 0;
    myHRLE->lower_hrle->max = 7;

    // 1. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 0;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 1;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );

    // 2. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 2;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );

    // 3. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 6;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );

    // 4. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 10;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );

    // 5. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 15;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );

    // 6. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 19;
    myHRLE->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->run_types.push_back( newRun );

    myHRLE->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->run_breaks.push_back( 5 );
    myHRLE->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->run_breaks.push_back( 5 );
    myHRLE->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->run_breaks.push_back( 6 );
    myHRLE->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->run_breaks.push_back( 6 );
    myHRLE->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->run_breaks.push_back( 6 );
    myHRLE->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->run_breaks.push_back( 5 );

    myHRLE->lower_hrle->start_indices.push_back( 0 );
    myHRLE->lower_hrle->start_indices.push_back( 5 );
    myHRLE->lower_hrle->start_indices.push_back( 8 );
    myHRLE->lower_hrle->start_indices.push_back( 11 );
    myHRLE->lower_hrle->start_indices.push_back( 14 );
    myHRLE->lower_hrle->start_indices.push_back( 17 );

    // dimension 1
    myHRLE->lower_hrle->lower_hrle->min = 0;
    myHRLE->lower_hrle->lower_hrle->max = 7;

    // 1. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 0;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 2. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 1;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 3. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 2;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 4. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 3;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 5. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 4;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 5;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 6. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 6;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 7. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 7;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 8. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 10;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 9. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 15;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 10. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 18;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 11. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 19;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 12. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 20;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 13. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 23;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 24;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 14. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 25;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 15. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 28;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 16. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 29;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 17. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 32;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 18. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 35;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 19. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 38;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 20. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 39;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 21. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 40;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    // 22. segment
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 41;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_DEFINED;
    newRun.index = 42;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );
    newRun.type = RT_POSITIVE;
    myHRLE->lower_hrle->lower_hrle->run_types.push_back( newRun );

    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );

    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );

    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 6 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );

    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );

    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );

    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 1 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 2 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->run_breaks.push_back( 4 );

    for (int i = 0; i < 2; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(2);
    for (int i = 0; i < 5; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(1);
    for (int i = 0; i < 6; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(0);
    for (int i = 0; i < 1; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(1);
    for (int i = 0; i < 1; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(2);
    for (int i = 0; i < 3; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(0);
    for (int i = 0; i < 2; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(1);
    for (int i = 0; i < 8; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(0);
    for (int i = 0; i < 1; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(1);
    for (int i = 0; i < 9; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(0);
    for (int i = 0; i < 5; i++) myHRLE->lower_hrle->lower_hrle->defined_values.push_back(1);

    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 0 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 3 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 6 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 9 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 12 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 17 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 20 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 23 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 26 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 29 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 32 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 35 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 38 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 43 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 46 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 49 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 52 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 55 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 58 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 61 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 64 );
    myHRLE->lower_hrle->lower_hrle->start_indices.push_back( 67 );

    this->bHrleObjectStatus = true;

    this->myHRLE->printDataStructure();

    this->bHrleObjectChanged = true;

//     this->myHRLE->printNeighborOfGridPoint( 1, 2, 2 , 0 );

    this->dynListsGL();
}

/////////////////////////////////////////////////////////////////////////////////////////
//          get and set methodes
/////////////////////////////////////////////////////////////////////////////////////////

bool CqOpenGLwidget::getColorMaterialStatus() const
{
  return bColorMaterial;
}


void CqOpenGLwidget::setColorMaterialStatus(bool theValue)
{
  bColorMaterial = theValue;
}


bool CqOpenGLwidget::getCoordinatesStatus() const
{
  return bCoordinates;
}


void CqOpenGLwidget::setCoordinatesStatus(bool theValue)
{
  bCoordinates = theValue;
}


bool CqOpenGLwidget::getCullBackFaceStatus() const
{
  return bCullBackFace;
}


void CqOpenGLwidget::setCullBackFaceStatus(bool theValue)
{
  bCullBackFace = theValue;
}


bool CqOpenGLwidget::getLightingStatus() const
{
  return bLighting;
}


void CqOpenGLwidget::setLightingStatus(bool theValue)
{
  bLighting = theValue;
}


bool CqOpenGLwidget::getTransparencyStatus() const
{
  return bTransparency;
}


void CqOpenGLwidget::setTransparencyStatus(bool theValue)
{
  bTransparency = theValue;
}


int CqOpenGLwidget::getCameraMode() const
{
  return iCameraMode;
}


void CqOpenGLwidget::setCameraMode(const int& theValue)
{
  iCameraMode = theValue;
}


bool CqOpenGLwidget::getWireFrameStatus() const
{
  return bWireFrame;
}


void CqOpenGLwidget::setWireFrameStatus(bool theValue)
{
  bWireFrame = theValue;
}


bool CqOpenGLwidget::getSmfObjectStatus() const
{
  return bSmfObjectStatus;
}


void CqOpenGLwidget::setSmfObjectStatus(bool theValue)
{
  bSmfObjectStatus = theValue;
}


bool CqOpenGLwidget::getVoxelObjectStatus() const
{
  return bVoxelObjectStatus;
}


void CqOpenGLwidget::setVoxelObjectStatus(bool theValue)
{
  bVoxelObjectStatus = theValue;
}


bool CqOpenGLwidget::getHrleObjectStatus() const
{
  return bHrleObjectStatus;
}


void CqOpenGLwidget::setHrleObjectStatus(bool theValue)
{
  bHrleObjectStatus = theValue;
}

bool CqOpenGLwidget::getDrawVoxelObject() const
{
  return bDrawVoxelObject;
}

void CqOpenGLwidget::setDrawVoxelObject(bool theValue)
{
  bDrawVoxelObject = theValue;
}

bool CqOpenGLwidget::getDrawSmfObject() const
{
  return bDrawSmfObject;
}

void CqOpenGLwidget::setDrawSmfObject(bool theValue)
{
  bDrawSmfObject = theValue;
}

bool CqOpenGLwidget::getDrawHrleSeq() const
{
  return bDrawHrleSeq;
}

void CqOpenGLwidget::setDrawHrleSeq(bool theValue)
{
  bDrawHrleSeq = theValue;
}

bool CqOpenGLwidget::getDrawHrleRand() const
{
  return bDrawHrleRand;
}

void CqOpenGLwidget::setDrawHrleRand(bool theValue)
{
  bDrawHrleRand = theValue;
}

bool CqOpenGLwidget::getDtGridObjectStatus() const
{
    return bDtGridObjectStatus;
}

void CqOpenGLwidget::setDtGridObjectStatus(bool theValue)
{
    bDtGridObjectStatus = theValue;
}

bool CqOpenGLwidget::getDrawDtRand() const
{
    return bDrawDtRand;
}

void CqOpenGLwidget::setDrawDtRand(bool theValue)
{
    bDrawDtRand = theValue;
}

bool CqOpenGLwidget::getDrawDtSeq() const
{
    return bDrawDtSeq;
}

void CqOpenGLwidget::setDrawDtSeq(bool theValue)
{
    bDrawDtSeq = theValue;
}

/*!
    \fn CqOpenGLwidget::getMemoryUsageHRLE()
 */
double CqOpenGLwidget::getMemoryUsageHRLE()
{
    if ( this->bHrleObjectStatus )
        return this->myHRLE->getMemorySize();

    return 0.0;
}

vector<neighbors> CqOpenGLwidget::getNeighbors( int x, int y, int z, int axis ) {

    vector<neighbors> returnValue;

    if ( this->bHrleObjectStatus )
        returnValue = this->myHRLE->neighborAccess( x, y, z, axis );

    return returnValue;

//     return this->myHRLE->neighborAccess( x, y, z, axis );
}

double CqOpenGLwidget::getAccessTimeRA() {

    if ( this->bHrleObjectStatus )
        return this->myHRLE->accessTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessAverageTimeRA() {

    if ( this->bHrleObjectStatus )
        return this->myHRLE->accessAverageTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessTimeSA() {

    if ( this->bHrleObjectStatus )
        return this->myHRLE->accessTimeSequentialAccess;

    return 0.0;

}

double CqOpenGLwidget::getMemoryUsageVoxelobject()
{
    if ( this->bVoxelObjectStatus )
        return this->myVoxelObject->getMemorySize();

    return 0.0;
}

double CqOpenGLwidget::getAccessTimeRA_voxelobject() {

    if ( this->bVoxelObjectStatus )
        return this->myVoxelObject->accessTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessAverageTimeRA_voxelobject() {

    if ( this->bVoxelObjectStatus )
        return this->myVoxelObject->accessAverageTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessTimeSA_voxelobject() {

    if ( this->bVoxelObjectStatus )
        return this->myVoxelObject->accessTimeSequentialAccess;

    return 0.0;

}

double CqOpenGLwidget::getMemoryUsageDtGrid()
{
    if ( this->bDtGridObjectStatus )
        return this->myDtGrid->getMemorySize();

    return 0.0;
}

double CqOpenGLwidget::getAccessTimeRA_dtgrid() {

    if ( this->bDtGridObjectStatus )
        return this->myDtGrid->accessTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessAverageTimeRA_dtgrid() {

    if ( this->bDtGridObjectStatus )
        return this->myDtGrid->accessAverageTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessTimeSA_dtgrid() {

    if ( this->bDtGridObjectStatus )
        return this->myDtGrid->accessTimeSequentialAccess;

    return 0.0;

}

bool CqOpenGLwidget::getPerfectHashObjectStatus() const
{
    return bPerfectHashObjectStatus;
}


void CqOpenGLwidget::setPerfectHashObjectStatus(bool theValue)
{
    bPerfectHashObjectStatus = theValue;
}


bool CqOpenGLwidget::getDrawPHRand() const
{
    return bDrawPHRand;
}


void CqOpenGLwidget::setDrawPHRand(bool theValue)
{
    bDrawPHRand = theValue;
}

double CqOpenGLwidget::getAccessTimeRA_pshash() {

    if ( this->bPerfectHashObjectStatus )
        return this->myPerfectHash->accessTimeRandomAccess;

// cout << "Perfect Hash Random Access Time: " << this->myPerfectHash->accessTimeRandomAccess << endl;

    return 0.0;

}

double CqOpenGLwidget::getAccessAverageTimeRA_pshash() {

    if ( this->bPerfectHashObjectStatus ) {
        return this->myPerfectHash->accessAverageTimeRandomAccess;

// cout << "Perfect Hash Random Access Average Time: " << this->myPerfectHash->accessAverageTimeRandomAccess << endl;
    }

    return 0.0;

}

double CqOpenGLwidget::getAccessTimeSA_pshash() {

    if ( this->bPerfectHashObjectStatus ) {
        return this->myPerfectHash->accessTimeSequentialAccess;

// cout << "Perfect Hash Sequential Access Time: " << this->myPerfectHash->accessTimeSequentialAccess << endl;
    }

    return 0.0;

}


double CqOpenGLwidget::getMemoryUsagePSHash()
{
    if ( this->bPerfectHashObjectStatus )
        return this->myPerfectHash->getMemorySize();

    return 0.0;
}

int CqOpenGLwidget::getNumDefinedVoxels() {
    if ( this->bVoxelObjectStatus )
        return this->myVoxelObject->numDefVoxels;

    return 0;
}

CPoint3D CqOpenGLwidget::getVoxelobjectResolution() {
    if ( this->bVoxelObjectStatus ) {
        return CPoint3D(
            this->myVoxelObject->m_ScopeX,
            this->myVoxelObject->m_ScopeY,
            this->myVoxelObject->m_ScopeZ );
    }

    return CPoint3D(0,0,0);
}

bool CqOpenGLwidget::getDrawSparseLevelSetRand() const
{
    return bDrawSparseLevelSetRand;
}


void CqOpenGLwidget::setDrawSparseLevelSetRand(bool theValue)
{
    bDrawSparseLevelSetRand = theValue;
}

bool CqOpenGLwidget::getDrawSparseLevelSetSeq() const
{
    return bDrawSparseLevelSetSeq;
}


void CqOpenGLwidget::setDrawSparseLevelSetSeq(bool theValue)
{
    bDrawSparseLevelSetSeq = theValue;
}


bool CqOpenGLwidget::getSparseLevelSetStatus() const
{
    return bSparseLevelSetStatus;
}


void CqOpenGLwidget::setSparseLevelSetStatus(bool theValue)
{
    bSparseLevelSetStatus = theValue;
}

double CqOpenGLwidget::getAccessTimeRA_sparse() {

    if ( this->bSparseLevelSetStatus )
        return this->mySparseLevelSet->accessTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessAverageTimeRA_sparse() {

    if ( this->bSparseLevelSetStatus )
        return this->mySparseLevelSet->accessAverageTimeRandomAccess;

    return 0.0;

}

double CqOpenGLwidget::getAccessTimeSA_sparse() {

    if ( this->bSparseLevelSetStatus )
        return this->mySparseLevelSet->accessTimeSequentialAccess;

    return 0.0;

}


double CqOpenGLwidget::getMemoryUsageSparse()
{
    if ( this->bSparseLevelSetStatus )
        return this->mySparseLevelSet->getMemorySize();

    return 0.0;
}


