/***************************************************************************
 *   Copyright (C) 2006 by Holger Gerth                                    *
 *   holger.gerth@informatik.tu-chemnitz.de                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "rlesparselevelset.h"

CRLESparseLevelSet::CRLESparseLevelSet()
{
}

CRLESparseLevelSet::CRLESparseLevelSet( CVoxelObject3D* vo )
{
    this->createRLEfromVoxelObject( vo );
}

CRLESparseLevelSet::~CRLESparseLevelSet()
{
}

run_code CRLESparseLevelSet::randomAccess( int X, int Y, int Z) {

    // create a stl vector from the given coordinates
    vector<int>     qv;
    qv.push_back(X);
    qv.push_back(Y);
    qv.push_back(Z);

    return this->randomAccess( qv );


}

run_code CRLESparseLevelSet::randomAccess( vector<int> qv) {
    unsigned int si = 0;                     // index into the segment table
    unsigned int firstRun = 0;               // first run of a segment
    int runPos = 0;                          // index into the run start table
    unsigned int segmentEnd;
    // run code for return value
    run_code returnValue;
    returnValue.type = RT_POSITIVE;
    // get the last coordinate
    int zCoord = qv.back();
    qv.pop_back();
    int yCoord = qv.back();
    qv.pop_back();
    int xCoord = qv.back();
    qv.pop_back();

    // find the correct segment where this grid point lies
    si = this->binSearchZ( zCoord, si, this->segmentStartTable.size() - this->numY);
    si = this->binSearchY( yCoord, si, si + this->numY -1 );

    // get the index into the run start table
    firstRun = this->segmentStartTable[si].index;

    // get the segment end
    if ((si+1) < this->segmentStartTable.size()) segmentEnd = this->segmentStartTable[si+1].index;
    else segmentEnd = this->runStartTable.size();

    // get the index into the run start array
    runPos = this->binSearch( xCoord, firstRun, segmentEnd-1 );

    if ( runPos >= 0 ) {
        returnValue.type = this->runStartTable[runPos].type;
        if ( returnValue.type == RT_DEFINED ) {
            returnValue.index = this->runStartTable[runPos].definedIndex + xCoord - this->runStartTable[runPos].xCoord;
        }
    }

    return returnValue;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//      Private functions
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*!
    \fn CRLESparseLevelSet::createRLEfromVoxelObject( CVoxelObject3D* vo )
 */
// convert an CVoxelObject3D into the HRLE structure
bool CRLESparseLevelSet::createRLEfromVoxelObject( CVoxelObject3D* vo ) {

    RUN_TYPE lastRun;
    this->maxX = vo->m_ScopeX;
    this->numY = vo->m_ScopeY;

    this->minRange = vo->m_MinRange;
    this->maxRange = vo->m_MaxRange;

    this->minBBox = CPoint3D(0,0,0);
    this->maxBBox = CPoint3D(vo->m_ScopeX, vo->m_ScopeY, vo->m_ScopeZ);

    // create RLE encoding
    for (int z = 0; z < vo->m_ScopeZ; z++) {
        for (int y = 0; y < vo->m_ScopeY; y++) {

            // create a new segment entry
            segmentCode newSegment;
            newSegment.index = this->runStartTable.size();
            newSegment.yCoord = y;
            newSegment.zCoord = z;

            this->segmentStartTable.push_back( newSegment );

            for (int x = 0; x < vo->m_ScopeX; x++) {

                // handle blocks which are outer
                if (vo->m_CubeArray(x,y,z).flag == CT_EMPTY) {
                    if ( (x == 0) || ( lastRun == RT_DEFINED ) ) {
                        startCode newRun;
                        newRun.xCoord = x;
                        newRun.type = RT_POSITIVE;
                        this->runStartTable.push_back( newRun );
                        lastRun = RT_POSITIVE;
                    }
                }

                // handle the border blocks
                else if (vo->m_CubeArray(x,y,z).flag == CT_BORDER) {
                    if ( (x == 0) || ( lastRun != RT_DEFINED ) ) {
                        startCode newRun;
                        newRun.xCoord = x;
                        newRun.type = RT_DEFINED;
                        newRun.definedIndex = this->value.size();
                        this->runStartTable.push_back( newRun );
                        lastRun = RT_DEFINED;
                    }
                    this->value.push_back(1.0);
                }

                // handle blocks which are inner
                else if (vo->m_CubeArray(x,y,z).flag == CT_INNER) {
                    if ( (x == 0) || ( lastRun == RT_DEFINED ) ) {
                        startCode newRun;
                        newRun.xCoord = x;
                        newRun.type = RT_NEGATIVE;
                        this->runStartTable.push_back( newRun );
                        lastRun = RT_NEGATIVE;
                    }
                }
            }
        }
    }
}

// binary search in the run start table
int CRLESparseLevelSet::binSearch( int xCoord, unsigned int start, unsigned int end ) {
    unsigned int mid = (end - start)/2 + start;

    if ( end < start ) return -1;

    if ( (this->runStartTable[mid].xCoord <= xCoord) && (( (mid < end) && (this->runStartTable[mid+1].xCoord > xCoord)) || ( (mid >= end) && (this->maxX > xCoord)) ) ) 
        return (int) mid;
    else if ( this->runStartTable[mid].xCoord < xCoord )
        return (int) this->binSearch( xCoord, mid+1, end );
    else 
        return (int) this->binSearch( xCoord, start, mid-1 );
}

unsigned int CRLESparseLevelSet::binSearchY( int yCoord, unsigned int start, unsigned int end ) {

    if (start == end) return start;

    unsigned int mid = (end - start)/2 + start;

    if (this->segmentStartTable[mid].yCoord == yCoord) return mid;
    else if ( this->segmentStartTable[mid].yCoord < yCoord ) return this->binSearchY( yCoord, mid+1, end);
    else return this->binSearchY( yCoord, start, mid-1 );
}

unsigned int CRLESparseLevelSet::binSearchZ( int zCoord, unsigned int start, unsigned int end ) {

    if (start == end) return start;

    unsigned int mid = ((end - start)/this->numY)/2 * this->numY + start;

    if (this->segmentStartTable[mid].zCoord == zCoord) return mid;
    else if ( this->segmentStartTable[mid].zCoord < zCoord ) return this->binSearchZ( zCoord, mid + this->numY, end);
    else return this->binSearchZ( zCoord, start, mid - this->numY );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
//              Testing
////////////////////////////////////////////////////////////////////////////////////////////////////////

// test the random access operator
void CRLESparseLevelSet::drawCubes()
{
    // for comnputing the average access time of the random access operator
    long            numGridPoints = 0, accessTime = 0, totalAccessTime = 0, averageAccessTime = 0;
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)

    int minX = this->minBBox.getX();
    int maxX = this->maxBBox.getX();
    int minY = this->minBBox.getY();
    int maxY = this->maxBBox.getY();
    int minZ = this->minBBox.getZ();
    int maxZ = this->maxBBox.getZ();

    float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
    float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
    float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

// cout << maxX << endl;
// cout << maxY << endl;
// cout << maxZ << endl;

    for (int z = minZ; z < maxZ; z++) {
        for (int y = minY; y < maxY; y++) {
            for (int x = minX; x < maxX; x++) {
                numGridPoints++;

// cout << "(" << x << ", " << y << ", " << z << ")" << endl;

                gettimeofday( &tp1, NULL );
                run_code run = this->randomAccess( x, y, z );
                gettimeofday( &tp2, NULL );

                accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);
                totalAccessTime += accessTime;

                if (run.type == RT_DEFINED) {

// cout << "Defined Value Index: " << run.index << endl;

                    // draw cubes and cube edges
                    // -------------------------

                    CVertex3D point1( this->minRange.getX() + x * CubeW,
                                        this->minRange.getY() + y * CubeH,
                                        this->minRange.getZ() + z * CubeD );
                    CVertex3D point2( this->minRange.getX() + (x+1) * CubeW,
                                        this->minRange.getY() + (y+1) * CubeH,
                                        this->minRange.getZ() + (z+1) * CubeD);

                    this->drawACube(&point1, &point2, this->value.at(run.index));

                }
            }
        }
    }


    this->accessTimeRandomAccess = (double) accessTime;
    this->accessAverageTimeRandomAccess = (double) totalAccessTime / (double) numGridPoints;
}

void CRLESparseLevelSet::drawCubesUsingSequentialAccess() {
    // for comnputing the average access time of the random access operator
    long            numGridPoints = 0, accessTime = 0, totalAccessTime = 0, averageAccessTime = 0;
    struct timeval  tp1,tp2;            // process cpu time in seconds (tv_sec) and micro seconds (tv_usec)

    int segment, runPos, segmentStart, segmentEnd, runStart, runEnd, xCoord, index;
    startCode run;

    int minX = this->minBBox.getX();
    int maxX = this->maxBBox.getX();
    int minY = this->minBBox.getY();
    int maxY = this->maxBBox.getY();
    int minZ = this->minBBox.getZ();
    int maxZ = this->maxBBox.getZ();

    float CubeW = (this->maxRange.getX() - this->minRange.getX()) / (float)(maxX - minX);
    float CubeH = (this->maxRange.getY() - this->minRange.getY()) / (float)(maxY - minY);
    float CubeD = (this->maxRange.getZ() - this->minRange.getZ()) / (float)(maxZ - minZ);

    gettimeofday( &tp1, NULL );
    for ( segment = 0; segment < this->segmentStartTable.size(); segment++) {

        segmentStart = this->segmentStartTable[segment].index;
        if ( (segment+1) < this->segmentStartTable.size() ) segmentEnd = this->segmentStartTable[segment+1].index;
        else segmentEnd = runStartTable.size();

        for ( runPos = segmentStart; runPos < segmentEnd; runPos++) {

            run = this->runStartTable.at(runPos);

            if (run.type == RT_DEFINED) {

                runStart = run.xCoord;
                if ( (runPos+1) < segmentEnd ) runEnd = this->runStartTable[runPos+1].xCoord;
                else runEnd = this->maxX;

                for ( xCoord = runStart; xCoord < runEnd; xCoord++) {
                    // draw cubes and cube edges
                    // -------------------------

                    CVertex3D point1( this->minRange.getX() + xCoord * CubeW,
                                        this->minRange.getY() + this->segmentStartTable[segment].yCoord * CubeH,
                                        this->minRange.getZ() + this->segmentStartTable[segment].zCoord * CubeD );
                    CVertex3D point2( this->minRange.getX() + (xCoord+1) * CubeW,
                                        this->minRange.getY() + (this->segmentStartTable[segment].yCoord+1) * CubeH,
                                        this->minRange.getZ() + (this->segmentStartTable[segment].zCoord+1) * CubeD);

                    index = run.definedIndex + xCoord - runStart;

                    this->drawACube(&point1, &point2, this->value.at(index));
                }

            }

        }
    }
    gettimeofday( &tp2, NULL );

    accessTime = (tp2.tv_sec -tp1.tv_sec) * 1000000 + (tp2.tv_usec - tp1.tv_usec);

//     this->accessTimeRandomAccess = (double) accessTime;
//     this->accessAverageTimeRandomAccess = (double) totalAccessTime / (double) numGridPoints;
    this->accessTimeSequentialAccess = (double) accessTime;
}

void CRLESparseLevelSet::drawACube(CVertex3D *p1, CVertex3D *p2, double colorValue) {

    // Fl�hen zeichnen
    // ----------------
        // Farbe festlegen
        glColor3f(0.5, 0.3, 1.0 * colorValue);

        // Offset fr den Cube (bessere Darstellung der Kanten)
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(1.f, 1.f);

            glBegin(GL_POLYGON);
                glNormal3f(  -1.0,  0.0,  0.0);
                glVertex3f(  p1->getX(), p1->getY(), p1->getZ());
                glVertex3f(  p1->getX(), p1->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p2->getZ());
                glVertex3f(  p1->getX(), p2->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d( 1.0,  0.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  -1.0,  0.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0, 1.0,  0.0);
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, -1.0);
                glVertex3d( p1->getX(), p1->getY(), p1->getZ());
                glVertex3d( p1->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p2->getY(), p1->getZ());
                glVertex3d( p2->getX(), p1->getY(), p1->getZ());
            glEnd();

            glBegin(GL_POLYGON);
                glNormal3d(  0.0,  0.0, 1.0);
                glVertex3d( p2->getX(), p1->getY(), p2->getZ());
                glVertex3d( p2->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p2->getY(), p2->getZ());
                glVertex3d( p1->getX(), p1->getY(), p2->getZ());
            glEnd();

        glDisable(GL_POLYGON_OFFSET_FILL);


// Kanten zeichnen
// ---------------

        glColor3d(0.2, 0.2, 0.2);

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p1->getY(), p1->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p1->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p1->getZ());
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p2->getX(), p2->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p2->getY(), p2->getZ());
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
        glEnd();

        glBegin(GL_LINES);
            glVertex3d(p1->getX(), p1->getY(), p2->getZ());
            glVertex3d(p2->getX(), p1->getY(), p2->getZ());
        glEnd();


    return;
}

void CRLESparseLevelSet::printDataStructure() {

    cout << "Ausgabe der Datenstruktur: " << endl;
    int i;
    cout << "Run Start Table (" << this->runStartTable.size() << "): ";
    for (i = 0; i < this->runStartTable.size(); i++) {
        if ( this->runStartTable[i].type == RT_DEFINED ) cout << this->runStartTable[i].xCoord << " d "<< this->runStartTable[i].definedIndex <<" | ";
        if ( this->runStartTable[i].type == RT_NEGATIVE ) cout << this->runStartTable[i].xCoord << " - | ";
        if ( this->runStartTable[i].type == RT_POSITIVE ) cout << this->runStartTable[i].xCoord << " + | ";
    }
    cout << endl;
    cout << "Segment Start Table (" << this->segmentStartTable.size() << "): ";
    for (i = 0; i < this->segmentStartTable.size(); i++) 
        cout << "(" << this->segmentStartTable[i].zCoord << ", " << this->segmentStartTable[i].yCoord << ", " << this->segmentStartTable[i].index << ") | ";
    cout << endl;

    cout << "Defined Values (" << this->value.size() << "): ";
    for (i = 0; i < this->value.size(); i++) cout << this->value.at(i) << " | ";
    cout << endl;

}

double CRLESparseLevelSet::getMemorySize()
{
    double returnValue = 0.0;

    returnValue += this->runStartTable.size() * sizeof(startCode);
    returnValue += this->segmentStartTable.size() * sizeof(segmentCode);
    returnValue += this->value.size() * sizeof(double);
    returnValue += 3* sizeof(int);

    return returnValue;
}